# -*- coding: utf-8 -*-
"""

@author: Erwan
"""

from __future__ import print_function, absolute_import, division, unicode_literals

import os
import sys
from os.path import dirname
import importlib

def getProjectRoot():
    ''' Return the full path of the project root'''

    return dirname(dirname(__file__))

def import_from_module(module, name):
    ''' Import object 'name' from module 'module' 
    raises AttributeError if name doesnt exist
    
    module, name: str
    
    '''
    impmodule = importlib.import_module(module)
    return getattr(impmodule , name)

class Chdir:
    ''' because we need to change directory to get into the NeQ folder to find
    the Git files and version, when imported from another program. This class
    then ensures we get back in the correct directory 
    
    Use
    -----
    
    > cd = Chdir(os.path.dirname(__file__))
    > try:
    >     (...)
    > except:
    >     (...)
    > finally:
    >     (...)
    >     cd.__del__()
    
    '''

    def __init__(self, newPath):
        self.savedPath = os.getcwd()
        os.chdir(newPath)

    def __del__(self):
        os.chdir(self.savedPath)

# %%
#==============================================================================
# Python 2/3 compatibility
#==============================================================================

if sys.version_info[0] == 2:
    import six.moves.configparser as configparser
    FileNotFoundError = OSError   # FileNotFoundError doesn't exist in Python 2....
    PermissionError = IOError   # PermissionError doesn't exist in Python 2....
else:
    import configparser
    FileNotFoundError = FileNotFoundError
    PermissionError = PermissionError
    
# %% Other stuff
    
class DatabankNotFound(FileNotFoundError):
    ''' Used when a line database is not found in neq.rc ''' 
    pass

# %%
#==============================================================================
# Optional packages
#==============================================================================

class NotInstalled(object):
    ''' A class to deal with optional packages 
    Will raise an error only if the package is used (but not if imported only)
    '''
    def __init__(self, name,info=''):
        self.__name = name
        self.__info = info

    def __getattr__(self,item):
        raise ImportError('The {0} package is required to use this '
                          'feature. {1}'.format(self.__name,self.__info))

    def __call__(self, *args, **kwargs):
        raise ImportError('The {0} package is required to use this '
                          'feature. {1}'.format(self.__name,self.__info))

# %% Test

def _test(*args, **kwargs):
    
    print('Project root:', getProjectRoot())
    
    return True

if __name__=='__main__':
    _test()
    



# -*- coding: utf-8 -*-
"""
Created on Sun May 07 20:34:22 2017

@author: p38

Description
-------

Signal processing functions 

"""

from __future__ import absolute_import, division, print_function, unicode_literals

from numpy import trapz, abs, linspace, isnan, nan
from scipy.interpolate import splev, splrep
from warnings import warn
from radis.misc.debug import printdbg


def resample(xspace, vector, xspace_new, k=1, ext='error', energy_threshold=1e-3,
             print_conservation=True):
    ''' Resample (xspace, vector) on a new space (xspace_new) of evenly distributed
    data and whose bounds are taken as the same as `xspace`. 
    
    Uses spline interpolation to create the intermediary points. Number of points
    is the same as the initial xspace, times a resolution factor. Verifies energy 
    conservation on the intersecting range at the end.
    
    
    Parameters    
    ----------
    
    xspace: array
        space on which vector was generated 
        
    vector: array
        quantity to resample
        
    resfactor: array
        xspace vector to resample on
    
    k: int
        order of spline interpolation. 3: cubic, 1: linear. Default 1.

    ext: 'error', 'extrapolate', 0, 1 
        Controls the value returned for elements of xspace_new not in the interval 
        defined by xspace. If 'error', raise a ValueError. If 'extrapolate', well,
        extrapolate. If '0' or 0, then fill with 0. If 1, fills with 1. 
        Default 'error'.

    energy_threshold: float
        if energy conservation (integrals on the intersecting range) is above 
        this threshold, raise an error. If None, dont check for energy conservation
        Default 1e-3 (0.1%)
        
    print_conservation: boolean
        if True, prints energy conservation


    Returns
    -------
    
    vector_new: array
        resampled vector on evenly spaced array. Number of element is conserved.
    
    Note that depending upon the from_space > to_space operation, sorting may 
    be reversed. 
    
    
    Examples
    --------
    
    >>> w_nm, I_nm = s.get('radiance')
    >>> w_cm, I_cm = resample_even(nm2cm(w_nm), I_nm)
    
    
    '''
    
    if len(xspace) != len(vector):
        raise ValueError('vector and xspace should have the same length. '+\
                         'Got {0}, {1}'.format(len(vector), len(xspace)))
    
    # Check reversed (interpolation requires objects are sorted)
    reverse = (xspace[-1] < xspace[0])
    
    if reverse:
        xspace = xspace[::-1]
        xspace_new = xspace_new[::-1]
        vector = vector[::-1]
    
    # translate ext in FITPACK syntax for splev
    if ext == 'extrapolate':
        ext_fitpack = 0          # splev returns extrapolated value
    elif ext in [0, '0', 1, '1', nan, 'nan']:
        ext_fitpack = 1          # splev returns 0  (fixed in post-processing)
    elif ext == 'error':
        ext_fitpack = 2          # splev raises ValueError
    else:
        raise ValueError('Unexpected value for `ext`: {0}'.format(ext))
        
    if isnan(vector).sum()>0:
        raise ValueError('Resampled vector has {0} nans. Interpolation will fail'.format
                         (isnan(vector).sum()))

    # Resample the slit function on the spectrum grid
    try:
        tck = splrep(xspace, vector, k=k)
    except ValueError:
        # Probably error on input data. Print it before crashing.
        print('\nValueError - Input data below:')
        print('-'*5)
        print(xspace)
        print(vector)
        print('Check plot 101 too')
        import matplotlib.pyplot as plt
        plt.figure(101).clear()
        plt.plot(xspace, vector)
        plt.xlabel('xspace')
        plt.ylabel('vector')
        plt.title('ValueError')
        raise
    vector_new = splev(xspace_new, tck, ext=ext_fitpack)

    # ... get masks
    b = (xspace >= xspace_new.min()) * (xspace <= xspace_new.max())
    b_new = (xspace_new >= xspace.min()) * (xspace_new <= xspace.max())
    
    # fix filling for out of boundary values
    if ext in [1, '1']:
        vector_new[~b_new] = 1
        if __debug__: printdbg('Filling with 1 on w<{0}, w>{1} ({2} points)'.format(
                xspace.min(), xspace.max(), (1-b_new).sum()))
    elif ext in [nan, 'nan']:
        vector_new[~b_new] = nan
        if __debug__: printdbg('Filling with nans on w<{0}, w>{1} ({2} points)'.format(
                xspace.min(), xspace.max(), (1-b_new).sum()))

    # Check energy conservation:

    # ... calculate energy
    energy0 = abs(trapz(vector[b], x=xspace[b]))
    energy_new = abs(trapz(vector_new[b_new], x=xspace_new[b_new]))
    if energy_new == 0: # deal with particular case of energy = 0
        if energy0 == 0:
            energy_ratio = 1
        else:
            energy_ratio = 0
    else:  # general case
        energy_ratio = energy0/energy_new
    if energy_threshold:
        if abs(energy_ratio-1) > energy_threshold:
            import matplotlib.pyplot as plt
            plt.figure(101).clear()
            plt.plot(xspace, vector, '-ok', label='original')
            plt.plot(xspace_new, vector_new, '-or', label='resampled')
            plt.xlabel('xspace')
            plt.ylabel('vector')
            plt.legend()
            raise ValueError('Energy conservation ({0:.5g}%) below tolerance level ({1:.5g}%)'.format(
                    (1-energy_ratio)*100, energy_threshold*100)+'. Check graph 101. '+\
                    'Increasing energy_threshold is possible but not recommended')
    if print_conservation:
        print('Resampling - Energy conservation: {0:.5g}%'.format(
                energy_ratio*100))
        
    # Reverse again
    if reverse:
        #xspace_new = xspace_new[::-1]
        vector_new = vector_new[::-1]
        
    return vector_new


def resample_even(xspace, vector, resfactor=2, k=1, ext='error', energy_threshold=1e-3,
             print_conservation=True):
    ''' Resample (xspace, vector) on a new space (xspace_new) of evenly distributed
    data and whose bounds are taken as the same as `xspace`. 
    
    Uses spline interpolation to create the intermediary points. Number of points
    is the same as the initial xspace, times a resolution factor. Verifies energy 
    conservation at the end.
    
    
    Parameters    
    ----------
    
    xspace: array
        space on which vector was generated 
        
    vector: array
        quantity to resample
        
    resfactor: float
        increase of resolution. If 1, output vector has the same number of 
        points as the input vector. Default 2. 
    
    k: int
        order of spline interpolation. 3: cubic, 1: linear. Default 1.

    ext: 'error', 'extrapolate', 0
        Controls the value returned for elements of xspace_new not in the interval 
        defined by xspace. If 'error', raise a ValueError. If 'extrapolate', well,
        extrapolate. If '0' or 0, then fill with 0. Default 'error'.

    energy_threshold: float
        if energy conservation (integrals) is above this threshold, raise an 
        error
        
    print_conservation: boolean
        if True, prints energy conservation


    Returns
    -------
    
    xspace_new: array
        evenly spaced mapping of xspace (same min, same max)
    
    vector_new: array
        resampled vector on evenly spaced array. Number of element is conserved.
    
    Note that depending upon the from_space > to_space operation, sorting may 
    be reversed. 
    
    
    Examples
    --------
    
    >>> w_nm, I_nm = s.get('radiance')
    >>> w_cm, I_cm = resample_even(nm2cm(w_nm), I_nm)
    
    
    '''
    
    # Get new evenly space array
    xspace_new = linspace(xspace[0], xspace[-1], len(vector)*resfactor)
    
    # Resample
    vector_new = resample(xspace, vector, xspace_new, k=k, ext=ext,
                          energy_threshold=energy_threshold,
                          print_conservation=print_conservation)
    
    return xspace_new, vector_new


# %% Test routines

def _test(verbose=True, debug=False, plot=True, warnings=True, *args, **kwargs):
    ''' Test procedures
    
    
    Parameters    
    ----------
    
    debug: boolean
        swamps the console namespace with local variables. Default False
        
    '''

    from neq.test.utils import getTestFile
    from radis.phys.convert import nm2cm, cm2nm
    import matplotlib.pyplot as plt
    from numpy import loadtxt, linspace

    # Test even resampling

    w_nm, I_nm = loadtxt(getTestFile('spectrum.txt')).T
    w_cm, I_cm = resample_even(nm2cm(w_nm), I_nm, resfactor=2, energy_threshold=1e-3,
                          print_conservation=verbose)
    
    if plot:
        plt.figure()             
        plt.xlabel('Wavelength (nm)')
        plt.ylabel('Intensity')
        plt.plot(w_nm, I_nm, '-ok', label='original')
        plt.plot(cm2nm(w_cm), I_cm, '-or', label='resampled')
        plt.legend()

    # Test resampling 
    
    w_crop = linspace(376, 381,100)    
    I_crop = resample(w_nm, I_nm, w_crop, energy_threshold=0.01)
    
    if plot:
        plt.figure() 
        plt.xlabel('Wavelength (nm)')
        plt.ylabel('Intensity')
        plt.plot(w_nm, I_nm, '-ok', label='original')
        plt.plot(w_crop, I_crop, '-or', label='resampled')
        plt.legend()

    if debug:
        globals().update(locals())
        
    if warnings:
        warn('Testing resampling: no quantitative tests defined yet')

    return True  # no standard tests yet 

if __name__ == '__main__':
    _test(debug=False)



# -*- coding: utf-8 -*-
"""
Created on Thu Aug 10 15:44:48 2017

@author: erwan

Constants and correlations for air 
"""


def air_index_dispersion(lbd):
    ''' Return air index dispersion as a function of wavelength 
    
    Parameters
    ----------
    
    lbd: array-like (µm)
        wavelength 
        
    
    References
    ----------
    
    P. E. Ciddor. "Refractive index of air: new equations for the visible and 
    near infrared", Appl. Optics 35, 1566-1573 (1996)
    
    https://refractiveindex.info/?shelf=other&book=air&page=Ciddor
    Standard air: dry air at 15 °C, 101.325 kPa and with 450 ppm CO2 content.
    
    '''
    
    n = 1 + 0.05792105/(238.0185-lbd**-2)+0.00167917/(57.362-lbd**-2)
    
    return n 


# Wavelength medium conversion functions

def vacuum2air(wavelength):
    ''' Converts wavelength as seen in vacuum to wavelength as seen in air
    
    
    Parameters    
    ----------
    
    wavelength: array-like (nm)
        wavelength in vacuum
        
    
    Returns
    -------
    
    wavelength: array-like (nm)
        wavelength in air
    
    '''
    
    air_index = air_index_dispersion(wavelength * 1e-3)      # nm > µm
    return wavelength / air_index

def air2vacuum(wavelength):
    ''' Converts wavelength as seen in air to wavelength as seen in vacuum
    
    
    Parameters    
    ----------
    
    wavelength: array-like (nm)
        wavelength in air
    
    
    Returns
    -------
    
    wavelength: array-like (nm)
        wavelength in vacuum
        
        
    Note
    ----
    
    Not exactly true, as air_index_dispersion is defined for vacuum wavelength
    However, air_index_dispersion doesnt vary much on 1-2 cm-1 (which is typical
    of air index dispersion effect on air/vacuum wavelength in the mid-IR)
    
    Estimate the error you make with:
        
    >>> vacuum2air(air2vacuum(w))-w
    
    '''
    
    air_index = air_index_dispersion(wavelength * 1e-3)       # nm > µm
    return air_index * wavelength



# -*- coding: utf-8 -*-
'''
HITRAN database parser 
'''

from __future__ import print_function, absolute_import, division, unicode_literals

import sys
import pandas as pd
import numpy as np
from collections import OrderedDict
import os
from os.path import exists, splitext
from six.moves import range
from six.moves import zip

# %% Hitran groups and classes
# As defined in Rothman et al, "The HITRAN 2004 molecular spectroscopic database"
# Tables 3 and 4

# Groups (define local quanta)

HITRAN_GROUP1 = ['H2O', 'O3', 'SO2', 'NO2', 'HNO3', 'H2CO', 'HOCl', 'H2O2', 'COF2',
               'H2S', 'HO2', 'HCOOH', 'ClONO2', 'HOBr', 'C2H4'] # asymmetric rotors

HITRAN_GROUP2 = ['CO2', 'N2O', 'CO', 'HF', 'HCl', 'HBr', 'HI', 'OCS', 'N2', 'HCN',
                 'C2H2', 'NO+']   # diatomic and linear molecules

HITRAN_GROUP3 = ['SF6', 'CH4']   # Spherical rotors

HITRAN_GROUP4 = ['CH3D', 'CH3Cl', 'C2H6', 'NH3', 'PH3', 'CH3OH']  # symmetric rotors

HITRAN_GROUP5 = ['O2']  # Triplet-Sigma ground electronic states

HITRAN_GROUP6 = ['NO', 'OH', 'ClO']  # Doublet-Pi ground electronic states 

# Classes (define global quanta)

HITRAN_CLASS1 = ['CO', 'HF', 'HCl', 'HBr', 'HI', 'N2', 'NO+']

HITRAN_CLASS2 = ['O2']  # Diatomic molecules with different electronic levels

HITRAN_CLASS3 = ['NO', 'OH', 'ClO']  # Diatomic molecules with doublet-Pi electronic state
    
HITRAN_CLASS4 = ['N2O', 'OCS', 'HCN']  # Linear triatomic

HITRAN_CLASS5 = ['CO2']  # Linear triatomic with large Fermi resonance

HITRAN_CLASS6 = ['H2O', 'O3', 'SO2', 'NO2', 'HOCl', 'H2S', 'HO2', 'HOBr'] # Non-linear triatomic

HITRAN_CLASS7 = ['C2H2']  # Linear tetratomic

HITRAN_CLASS8 = ['NH3', 'PH3']  # Pyramidal tetratomic

HITRAN_CLASS9 = ['H2CO', 'H2O2', 'COF2']  # Non-linear tetratomic

HITRAN_CLASS10 = ['CH4', 'CH3D', 'CH3Cl', 'C2H6', 'HNO3', 'SF6', 'HCOOH', 
                  'ClONO2', 'C2H4', 'CH3OH']  # Pentatomic or greater polyatomic

# %% Parsing functions

# General case
columns_2004 = OrderedDict([ (
               # name    # format # type  # description                                 # unit 
               'id',     ('a2',   int,   'Molecular number'                               ,''                      )),(
               'iso',    ('a1',   int,   'isotope number'                                 ,''                      )),(
               'wav',    ('a12',  float, 'vacuum wavenumber'                              ,'cm-1'                  )),(
               'int',    ('a10',  float, 'intensity at 296K'                              ,'cm-1/(molecule/cm-2)', )),(
               'A',      ('a10',  float, 'Einstein A coefficient'                         ,'s-1'                   )),(
               'airbrd', ('a5',   float, 'air-broadened half-width at 296K'               ,'cm-1.atm-1'            )),(
               'selbrd', ('a5',   float, 'self-broadened half-width at 296K'              ,'cm-1.atm-1'            )),(
               'El',     ('a10',  float, 'lower-state energy'                             ,'cm-1'                  )),(
               'Tdpair', ('a4',   float, 'temperature-dependance exponent for Gamma air'  ,''                      )),(
               'Pshft',  ('a8',   float, 'air pressure-induced line shift at 296K'        ,'cm-1.atm-1'            )),( 
               'globu',  ('a15',  str,   'electronic and vibrational global upper quanta' ,''                      )),(
               'globl',  ('a15',  str,   'electronic and vibrational global lower quanta' ,''                      )),(
               'locu',   ('a15',  str,   'electronic and vibrational local upper quanta'  ,''                      )),(
               'locl',   ('a15',  str,   'electronic and vibrational local lower quanta'  ,''                      )),(
               'ierr',   ('a6',   str,   'ordered list of indices corresponding to uncertainty estimates of transition parameters'    ,''   )),(
               'iref',   ('a12',  str,   'ordered list of reference identifiers for transition parameters'                    ,''           )),(
               'lmix',   ('a1',   str,   'flag indicating the presence of additional data and code relating to line-mixing'   ,''           )),(
               'gp',     ('a7',   float, 'upper state degeneracy'                         ,''                      )),(
               'gpp',    ('a7',   float, 'lower state degeneracy'                         ,''                      ))
               ])


# Particular cases for different groups of local quanta ('locu', 'locl' are split)
# cf "The HITRAN 2004 molecular spectroscopic database", Rothman et al. 2004
# -----------

# Group 2: diatomic and linear molecules
columns_grp2 = OrderedDict([ (
               # name    # format # type  # description                                 # unit 
               'id',     ('a2',   int,   'Molecular number'                               ,''                      )),(
               'iso',    ('a1',   int,   'isotope number'                                 ,''                      )),(
               'wav',    ('a12',  float, 'vacuum wavenumber'                              ,'cm-1'                  )),(
               'int',    ('a10',  float, 'intensity at 296K'                              ,'cm-1/(molecule/cm-2)', )),(
               'A',      ('a10',  float, 'Einstein A coefficient'                         ,'s-1'                   )),(
               'airbrd', ('a5',   float, 'air-broadened half-width at 296K'               ,'cm-1.atm-1'            )),(
               'selbrd', ('a5',   float, 'self-broadened half-width at 296K'              ,'cm-1.atm-1'            )),(
               'El',     ('a10',  float, 'lower-state energy'                             ,'cm-1'                  )),(
               'Tdpair', ('a4',   float, 'temperature-dependance exponent for Gamma air'  ,''                      )),(
               'Pshft',  ('a8',   float, 'air pressure-induced line shift at 296K'        ,'cm-1.atm-1'            )),( 
               'globu',  ('a15',  str,   'electronic and vibrational global upper quanta' ,''                      )),(
               'globl',  ('a15',  str,   'electronic and vibrational global lower quanta' ,''                      )),(
               #'locu',   ('a15',  str,   'electronic and vibrational local upper quanta'  ,''                      )),(
                # 10X included in Fu
               'Fu',     ('a15',  str,  'upper state total angular momentum including nuclear spin'  ,''         )),(
               #'locl',   ('a15',  str,   'electronic and vibrational local lower quanta'  ,''                      )),(
               'branch', ('a6',   str,     'O, P, Q, R, S branch symbol'                  ,''                      )),(
               'jl',     ('a3',   int,    'lower state rotational quantum number'         ,''                      )),(
               'sym',    ('a1',   str,     'symmetry'                                     ,''                      )),(
               'Fl',     ('a5',   str,   'lower state total angular momentum including nuclear spin', ''         )),(
               #
               'ierr',   ('a6',   str,   'ordered list of indices corresponding to uncertainty estimates of transition parameters'   ,''    )),(
               'iref',   ('a12',  str,   'ordered list of reference identifiers for transition parameters'                    ,''           )),(
               'lmix',   ('a1',   str,   'flag indicating the presence of additional data and code relating to line-mixing'   ,''           )),(
               'gp',     ('a7',   float, 'upper state degeneracy'                         ,''                      )),(
               'gpp',    ('a7',   float, 'lower state degeneracy'                         ,''                      ))
               ])

# quick fix # TODO: proper implementation of HITRAN classes, and groups
# Update: classes are now implemented properly, groups remain to be done 

# %% Hitran global quanta classes

def parse_HITRAN_class1(df):
    ''' Diatomic molecules: CO, HF, HCl, HBr, HI, N2, NO+
    
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    
    Notes
    -----
    
    HITRAN syntax:
    
    >>>      v1
    >>>  13x I2
    
    '''
    dgu = df['globu'].str.extract(
            '[ ]{13}(?P<v1u>[\d ]{2})',
            expand=True)
    dgl = df['globl'].str.extract(
            '[ ]{13}(?P<v1l>[\d ]{2})',
            expand=True)
    dgu = dgu.apply(pd.to_numeric)
    dgl = dgl.apply(pd.to_numeric)
    return pd.concat([df, dgu, dgl], axis=1)

def parse_HITRAN_class2(df):
    ''' Diatomic molecules with different electronic levels: O2
    
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    
    Notes
    -----
    
    HITRAN syntax:
    
    >>>      X  v1
    >>>  12x A1 I2
    
    '''
    raise NotImplementedError

def parse_HITRAN_class3(df):
    ''' Diatomic molecules with doublet-Pi electronic state: NO, OH, ClO
    
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    
    Notes
    -------
    
    HITRAN syntax:
    
    >>>      X i     v1
    >>>  7x A1 A3 2x I2
    '''
    raise NotImplementedError

def parse_HITRAN_class4(df):
    ''' Linear triatomic: N2O, OCS, HCN
    
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    
    Notes
    -----
    
    HITRAN syntax:
    
    >>>     v1 v2 l2 v3
    >>>  7x I2 I2 I2 I2
    
    Note: I2 in regexp: [\d ]{2} 
    '''
    dgu = df['globu'].str.extract(
            '[ ]{7}(?P<v1u>[\d ]{2})(?P<v2u>[\d ]{2})(?P<l2u>[\d ]{2})(?P<v3u>[\d ]{2})',
            expand=True)
    dgl = df['globl'].str.extract(
            '[ ]{7}(?P<v1l>[\d ]{2})(?P<v2l>[\d ]{2})(?P<l2l>[\d ]{2})(?P<v3l>[\d ]{2})',
            expand=True)
    dgu = dgu.apply(pd.to_numeric)
    dgl = dgl.apply(pd.to_numeric)
    return pd.concat([df, dgu, dgl], axis=1)

def parse_HITRAN_class5(df):
    ''' Linear triatomic with large Fermi resonance: CO2
    
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    
    Notes
    -----
    
    HITRAN syntax:
    
    >>>     v1 v2 l2 v3 r
    >>>  6x I2 I2 I2 I2 I1
    
    Note: I2 in regexp: [\d ]{2} 
    '''
    dgu = df['globu'].str.extract(
            '[ ]{6}(?P<v1u>[\d ]{2})(?P<v2u>[\d ]{2})(?P<l2u>[\d ]{2})(?P<v3u>[\d ]{2})(?P<ru>\d)',
            expand=True)
    dgl = df['globl'].str.extract(
            '[ ]{6}(?P<v1l>[\d ]{2})(?P<v2l>[\d ]{2})(?P<l2l>[\d ]{2})(?P<v3l>[\d ]{2})(?P<rl>\d)',
            expand=True)
    dgu = dgu.apply(pd.to_numeric)
    dgl = dgl.apply(pd.to_numeric)
    return pd.concat([df, dgu, dgl], axis=1)

def parse_HITRAN_class6(df):
    ''' Non-linear triatomic: H2O, O3, SO2, NO2, HOCl, H2S, HO2, HOBr
    
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    
    Notes
    -----
    
    HITRAN syntax:
    
    >>>     v1 v2 v3
    >>>  9x I2 I2 I2
    
    Note: I2 in regexp: [\d ]{2} 
    '''
    dgu = df['globu'].str.extract(
            '[ ]{9}(?P<v1u>[\d ]{2})(?P<v2u>[\d ]{2})(?P<v3u>[\d ]{2})',
            expand=True)
    dgl = df['globl'].str.extract(
            '[ ]{9}(?P<v1l>[\d ]{2})(?P<v2l>[\d ]{2})(?P<v3l>[\d ]{2})',
            expand=True)
    dgu = dgu.apply(pd.to_numeric)
    dgl = dgl.apply(pd.to_numeric)
    return pd.concat([df, dgu, dgl], axis=1)

def parse_HITRAN_class7(df):
    ''' Linear tetratomic: C2H2
    
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    
    Notes
    -----
    
    HITRAN syntax:
    
    >>>
    '''
    raise NotImplementedError

def parse_HITRAN_class8(df):
    ''' Pyramidal tetratomic: NH3, PH3
    
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    
    Notes
    -----
    
    HITRAN syntax:
    
    >>>
    '''
    raise NotImplementedError

def parse_HITRAN_class9(df):
    ''' Non-linear tetratomic: H2CO, H2O2, COF2
    
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    
    Notes
    -----
    
    HITRAN syntax:
    
    >>>
    '''
    raise NotImplementedError

def parse_HITRAN_class10(df):
    ''' Pentatomic or greater polyatomic
    
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    
    Notes
    -----
    
    HITRAN syntax:
    
    >>>
    '''
    raise NotImplementedError

# %% Reading function


def _format_dtype(dtype):
    ''' Format dtype from specific columns. Crash with hopefully helping error message '''
    
    try:
        dt = np.dtype([(str(k), c) for k, c in dtype])
        # Note: dtype names cannot be `unicode` in Python2. Hence the str()
    except TypeError:
        # Cant read database. Try to be more explicit for user
        print('Data type')
        print('-'*30)
        for (k, c) in dtype:
            print(str(k), '\t', c)
        print('-'*30)
        raise
    return dt

def _cast_to_dtype(data, dtype):
    ''' Cast array to certain type, crash with hopefull helping error message.
    Return casted data
    
    
    Parameters    
    ----------
    
    data: array to cast
    
    dtype: (ordered) list of (param, type)
    
    '''
    
    dt = _format_dtype(dtype)
    
    try:
        data = np.array(data, dtype=dt)
    except ValueError:
        try:
            # Cant read database. Try to be more explicit for user
            print('Cant cast data to specific dtype. Trying column by column:')
            print('-'*30)
            for i in range(len(data[0])):
                print(dtype[i], '\t', np.array(data[0][i], dtype=dt[i]))
            print('-'*30)
        except ValueError:
            print('>>> Next param:', dtype[i], '. Value:', data[0][i], '\n')
            raise ValueError('Cant cast data to specific dtype. Tried column by column. See results above')

    return data

def hit2df(fname, count=-1, cache=False, verbose=True):
    ''' Convert a HITRAN/HITEMP file to a Pandas dataframe 
    
    
    Parameters    
    ----------
    
    fname: str
        HITRAN-HITEMP file name 
        
    count: int
        number of items to read (-1 means all file)
        
    cache: boolean
        if True, a pandas-readable HDF5 file is generated on first access, 
        and later used. This saves on the datatype cast and conversion and
        improves performances a lot (but changes in the database are not 
        taken into account). If False, no database is used. If 'regen', temp
        file are reconstructed. Default False. 
    
    
    Returns
    -------
    
    df: pandas Dataframe
        dataframe containing all lines and parameters
        
    
    
    References
    ----------
    
    HITRAN-HITEMP doc 
    
    
    Notes
    -----
    
    Performances: see CDSD-HITEMP parser
    
    '''
    
    columns = columns_2004

    if cache: # lookup if cached file exist. 
#        fcache = fname+'.cached'
        fcache = splitext(fname)[0]+'.h5'
        if exists(fcache):
            if cache == 'regen':
                os.remove(fcache)
                if verbose: print('Deleted h5 cache file : {0}'.format(fcache))
            else:
                if verbose: print('Using h5 file: {0}'.format(fcache))
    #            return pd.read_csv(fcache)
                return pd.read_hdf(fcache, 'df')
        
    # Detect the molecule by reading the start of the file
    with open(fname) as f:
        mol = get_molecule(int(f.read(2)))
        
    # parse group specific local quanta information
    if mol in HITRAN_GROUP1:
        if verbose: print('Local quanta specific format not implemented yet for group 1') #TODO someday
    elif mol in HITRAN_GROUP2:
        columns = columns_grp2  
    elif mol in HITRAN_GROUP3:
        if verbose: print('Local quanta specific format not implemented yet for group 3') #TODO someday
    elif mol in HITRAN_GROUP4:
        if verbose: print('Local quanta specific format not implemented yet for group 4') #TODO someday
    elif mol in HITRAN_GROUP5:
        if verbose: print('Local quanta specific format not implemented yet for group 5') #TODO someday
    elif mol in HITRAN_GROUP6:
        if verbose: print('Local quanta specific format not implemented yet for group 6') #TODO someday
    else:
        if verbose: print('Unknown molecule group for:', mol)


    # %% Start reading the full file
    
    # get format of line return
    if sys.platform in ['linux', 'darwin']:
        linereturnformat = 'a1'
    elif sys.platform in ['win32','linux2']:
        linereturnformat = 'a2'
    else:
        raise ValueError('Line return format not defined for this OS: please update neq')

    # ... Create a dtype with the binary data format and the desired column names
    dtype = [(k, c[0]) for (k, c) in columns.items()]+[('_linereturn',linereturnformat)]   
    # ... _linereturn is to capture the line return symbol. We delete it afterwards
    dt = _format_dtype(dtype)
    data = np.fromfile(fname, dtype=dt, count=count)
    
    # ... Cast to new type
    # This requires to recast all the data already read, but is still the fastest
    # method I found to read a file directly (for performance benchmark see 
    # CDSD-HITEMP parser)
    newtype = [c[0] if (c[1]==str) else c[1] for c in columns.values()]
    dtype = list(zip(list(columns.keys()), newtype))+[('_linereturn',linereturnformat)]
    data = _cast_to_dtype(data, dtype)

    # %% Create dataframe    
    df = pd.DataFrame(data.tolist(), columns=list(columns.keys())+['_linereturn'])

    # assert one molecule per database only. Else the groupbase data reading 
    # above doesnt make sense
    nmol = len(set(df['id']))
    if nmol == 0:
        raise ValueError('Databank looks empty')        
    elif nmol!=1:
        raise ValueError('Multiple molecules in database ({0}). Current '.format(nmol)+\
                         'spectral code only computes 1 species at the time. Use MergeSlabs. '+\
                         'Verify the parsing was correct by looking at the first row below: '+\
                         '\n{0}'.format(df.iloc[0]))
    
    for k, c in columns.items():
        if c[1] == str:
            df[k] = df[k].str.decode("utf-8")
            
    # Add global quanta attributes
    if mol in HITRAN_CLASS1:
        df = parse_HITRAN_class1(df)
    elif mol in HITRAN_CLASS2:
        df = parse_HITRAN_class2(df)
    elif mol in HITRAN_CLASS3:
        df = parse_HITRAN_class3(df)
    elif mol in HITRAN_CLASS4:
        df = parse_HITRAN_class4(df)
    elif mol in HITRAN_CLASS5:
        df = parse_HITRAN_class5(df)
    elif mol in HITRAN_CLASS6:
        df = parse_HITRAN_class6(df)
    elif mol in HITRAN_CLASS7:
        df = parse_HITRAN_class7(df)
    elif mol in HITRAN_CLASS8:
        df = parse_HITRAN_class8(df)
    elif mol in HITRAN_CLASS9:
        df = parse_HITRAN_class9(df)
    elif mol in HITRAN_CLASS10:
        df = parse_HITRAN_class10(df)
    else:
        raise ValueError('Unknown class for molecule {0}. Cant parse global quanta'.format(
                mol))
            
    # Strip whitespaces around PQR columns (due to 2 columns jumped)
    if 'branch' in df:
        df['branch'] = df.branch.str.strip()
    
    # Delete dummy column than handled the line return character
    del df['_linereturn']
    
    if cache: # cached file mode but cached file doesn't exist yet (else we had returned)
        if verbose: print('Generating cached file: {0}'.format(fcache))
        try:
#            df.to_csv(fcache)
            df.to_hdf(fcache, 'df', format='fixed')
        except:
            if verbose:
                print(sys.exc_info())
                print('An error occured in cache file generation. Lookup access rights')
            pass
        
    return df 

def get_molecule_identifier(molecule_name):
    '''
    For a given input molecular formula, return the corresponding HITRAN molecule identifier number.
    
    
    Parameters
    ----------
    molecular_formula : str
        The string describing the molecule.
        
        
    Returns
    -------
    M : int
        The HITRAN molecular identified number.
        
        
    References
    ----------
    
    Function from https://github.com/nzhagen/hitran/blob/master/hitran.py
    '''

    trans = { '1':'H2O',    '2':'CO2',   '3':'O3',      '4':'N2O',   
              '5':'CO',    '6':'CH4',   '7':'O2',     '8':'NO',
              '9':'SO2',   '10':'NO2',  '11':'NH3',    '12':'HNO3', 
              '13':'OH',   '14':'HF',   '15':'HCl',   '16':'HBr',
             '17':'HI',    '18':'ClO',  '19':'OCS',    '20':'H2CO', 
             '21':'HOCl', '22':'N2',   '23':'HCN',   '24':'CH3Cl',
             '25':'H2O2',  '26':'C2H2', '27':'C2H6',   '28':'PH3',  
             '29':'COF2', '30':'SF6',  '31':'H2S',   '32':'HCOOH',
             '33':'HO2',   '34':'O',    '35':'ClONO2', '36':'NO+',  
             '37':'HOBr', '38':'C2H4', '39':'CH3OH', '40':'CH3Br',
             '41':'CH3CN', '42':'CF4',  '43':'C4H2',   '44':'HC3N', 
             '45':'H2',   '46':'CS',   '47':'SO3'}
    ## Invert the dictionary.
    trans = {v:k for k,v in trans.items()}

    return(int(trans[molecule_name]))

def get_molecule(molecule_id):
    '''
    For a given input molecular identifier, return the corresponding HITRAN 
    molecule name.
    
    
    Parameters    
    ----------
    
    molecular_id : str
        Hitran identifier of the molecule.
    '''
    
    # assert str
    id = '{0}'.format(molecule_id)

    trans = { '1':'H2O',    '2':'CO2',   '3':'O3',      '4':'N2O',   
              '5':'CO',    '6':'CH4',   '7':'O2',     '8':'NO',
              '9':'SO2',   '10':'NO2',  '11':'NH3',    '12':'HNO3', 
              '13':'OH',   '14':'HF',   '15':'HCl',   '16':'HBr',
             '17':'HI',    '18':'ClO',  '19':'OCS',    '20':'H2CO', 
             '21':'HOCl', '22':'N2',   '23':'HCN',   '24':'CH3Cl',
             '25':'H2O2',  '26':'C2H2', '27':'C2H6',   '28':'PH3',  
             '29':'COF2', '30':'SF6',  '31':'H2S',   '32':'HCOOH',
             '33':'HO2',   '34':'O',    '35':'ClONO2', '36':'NO+',  
             '37':'HOBr', '38':'C2H4', '39':'CH3OH', '40':'CH3Br',
             '41':'CH3CN', '42':'CF4',  '43':'C4H2',   '44':'HC3N', 
             '45':'H2',   '46':'CS',   '47':'SO3'}

    return(trans[id])

## ======================================================
# %% Test

def _test(verbose=True, warnings=True, **kwargs):
    ''' Analyse some default files to make sure everything still works'''
    from neq.test.utils import getTestFile
    from time import time
    
    b = True
    
    t0 = time()
    df = hit2df(getTestFile('hitran_CO_fragment.par'))
    print('File loaded in {0:.0f}s'.format(time()-t0))
    if verbose: print(df.head())
    b *= (list(df.ix[0, ['v1u', 'v1l']]) == [4, 4])

    t0 = time()
    df = hit2df(getTestFile('hitran_CO2_fragment.par'))
    print('File loaded in {0:.0f}s'.format(time()-t0))
    if verbose: print(df.head())
    b *= (list(df.ix[0, ['v1u', 'v2u', 'l2u', 'v3u', 'v1l', 'v2l', 'l2l', 'v3l']]) ==
              [4, 0, 0, 0, 0, 0, 0, 1])
    
    return (bool(b))


if __name__ == '__main__':
    print('Testing HITRAN parsing: ', _test())
        


# -*- coding: utf-8 -*-
"""
Created on Mon May 29 21:50:56 2017

@author: erwan

A Spectrum Database class to manage them all

It basically manages a list of Spectrum JSON files, adding a Pandas
dataframe structure on top to serve as an efficient index to visualize
the spectra input conditions, and slice through the Dataframe with
easy queries

Example
-------

    >>> from neq.spec import SpecDatabase
    >>> db = SpecDatabase(r"path/to/database")     # create or loads database

    >>> db.update()  # in case something changed (like a file was added manually)
    >>> db.see(['Tvib', 'Trot'])   # nice print in console

    >>> s = db.get('Tvib==3000 & Trot==1500')[0]  # get all spectra that fit conditions
    >>> db.add(s)  # update database (and raise error because duplicate!)

Note that SpectrumFactory objects can be configured to automatically update
a database

Edit database: 
    
An example of script to update all spectra conditions in a database (ex: when 
a condition was added afterwards to the Spectrum class)
    
    >>> # Example: add the 'medium' key in conditions 
    >>> db = "database_CO"
    >>> for f in os.listdir(db):
    >>>    if not f.endswith('.spec'): continue
    >>>    s = load_spec(join(db, f))
    >>>    s.conditions['medium'] = 'vacuum'
    >>>    s.store(join(db,f), if_exists_then='replace')

Todo
----

Alert if case already in database when generating from a SpectrumFactory / ParallelFactory
connected to a SpecDatabase

Implement a h5py version of load / store 

"""

from __future__ import absolute_import
from __future__ import print_function
import json
import numpy as np
from numpy import array
import pandas as pd
import matplotlib.pyplot as plt
from scipy.interpolate import griddata
import os
import sys
from warnings import warn
from os.path import (join, splitext, exists, basename, split, dirname, abspath, 
                     isdir, getsize)
from six.moves import range
from radis.spectrum.spectrum import Spectrum
from shutil import copy2
from time import strftime
from radis.misc.basics import is_float, list_if_float
from radis.misc.utils import FileNotFoundError, PermissionError
from radis.misc.debug import printdbg
from six import string_types


# Serializing functions
# ... functions to store / load a Spectrum to / from a JSON file

# %% Save functions

def is_jsonable(x):
    try:
        json.dumps(x)
        return True
    except:
        return False
    
def save(s, path, discard=[], compress=False, add_info=None, add_date=None, 
         if_exists_then='increment', verbose=True, warnings=True):
    ''' Save a Spectrum object in JSON format. Object can be recovered with 
    radis.tools.load_spec()
    
    Parameters
    ----------

    s: Spectrum
        to save

    path: str
        filename to save. No extension needed. If filename already
        exists then a digit is added. If filename is a directory then a new
        file is created within this directory.

    discard: list of str
        parameters to discard. To save some memory.
        
    compress: boolean
        if True, removes all quantities that can be regenerated with s.update(),
        e.g, transmittance if abscoeff and path length are given, radiance if
        emisscoeff and abscoeff are given in non-optically thin case, etc.
        Default False

    add_info: list, or None/False
        append these parameters and their values if they are in conditions.
        e.g:
        >>> add_info = ['Tvib', 'Trot']

    add_date: str, or None/False
        adds date in strftime format to the beginning of the filename.
        e.g:
        >>> add_date = '%Y%m%d'
        Default ''

    if_exists_then: 'increment', 'replace', 'error'
        what to do if file already exists. If increment an incremental digit
        is added. If replace file is replaced (yeah). If error (or anything else)
        an error is raised. Default `increment`

    Returns
    -------

    Returns filename used (may be different from given path as new info or
    incremental identifiers are added)

    '''

    # 1) Format to JSON writable dictionary
    sjson, conditions = _format_to_jsondict(s, discard, compress)

    # 2) Get final output name (add info, extension, increment number if needed)
    fout = _get_fout_name(path, if_exists_then, add_date, add_info, conditions, 
                          sjson, verbose)
    
    # 3) Now is time to save
    with open(fout, 'w') as f:
        try:
            json.dump(sjson, f)
        except TypeError:
            _export_safe(sjson, f, warnings=warnings)
    if verbose: print('Spectrum stored in {0} ({1:.1f}Mb)'.format(fout,
                       getsize(fout)/1024*1e-3))

    return fout    # return final name

def _format_to_jsondict(s, discard, compress):
    ''' Format to JSON writable dictionary '''
    
    # ... add main attributes from Spectrum class
    sjson = {}
    sjson['q'] = {}
    sjson['q_conv'] = {}
    for k, v in s._q.items():
        sjson['q'][k] = v.tolist()
    for k, v in s._q_conv.items():
        sjson['q_conv'][k] = v.tolist()
        
    if compress:
        sjson = _compress(s, sjson) #, s.conditions, equilibrium=s.is_at_equilibrium())
    
    # ... Format conditions, check that minimum information is given
    try:
        conditions = s.conditions
    except AttributeError:
        raise ValueError('Spectrum needs a `conditions` attribute (dict) to be stored in database')
    try:
        conditions['waveunit']
    except KeyError:
        raise KeyError('Spectrum `conditions` dict should at least have a `waveunit` key')
    # Todo: what if conditions is an empty dictionary? do we allow that?
    sjson['conditions'] = conditions

    # ... Only `quantities` and `conditions` are required. The rest is just extra
    # details. Add them now if they exist (assuming a Spectrum class is being stored)
    for attr in ['units', 'cond_units', 'name']:
        if attr not in discard:
            try:
                sjson[attr] = s.__getattribute__(attr)
            except AttributeError:
                pass
    # ... special case of lines (a Pandas dataframe): 
    if 'lines' not in discard:
        try:
            if s.lines is not None:
                sjson['lines'] = s.lines.to_json()   # Pandas > JSON conversion
                sjson['_dtypes_lines'] = s.lines.dtypes.to_json()  # Keep dtype 
        except AttributeError:
            pass  # dont store lines if they dont exist
    # ... special case of populations (a dict of dict of dict of Pandas dataframes)
    if 'populations' not in discard:
        try:
            if s.populations is not None:
                pops = s.populations
                jsonpops = {}
                for molecule, isotopes in pops.items():
                    jsonpops[molecule] = {}
                    for isotope, elec_states in isotopes.items():
                        jsonpops[molecule][isotope] = {}
                        for elec_state, content in elec_states.items():
                            jsonpops[molecule][isotope][elec_state] = {}
                            for k, v in content.items():
                                if isinstance(v, pd.DataFrame): # rovib or vib levels
                                    vjson = v.to_json()   # Pandas > JSON conversion
                                    jsonpops[molecule][isotope][elec_state][k] = vjson
                                    jsonpops[molecule][isotope][elec_state]['_dtypes_'+k] = v.dtypes.to_json()
                                else:   # can be other data. Hopefully serializable. 
                                    vjson = v
                                    jsonpops[molecule][isotope][elec_state][k] = vjson
            sjson['populations'] = jsonpops
        except AttributeError:
            pass  # dont store populations if they dont exist
    
    return sjson, conditions

def _get_fout_name(path, if_exists_then, add_date, add_info, conditions, sjson, verbose):
    ''' Get final output name   (add info, extension, increment number if needed) '''
    
    if isdir(path):
        fold, name = path, ''
    else:
        fold, name = split(path)

    # ... add date info
    if add_date not in ['', None, False]:
        date = strftime(add_date)
    else:
        date = ''

    # ... add conditions info
    if add_info not in [[], {}, None, False]:
        # complete name with info about calculation conditions
        info = []
        for k in add_info:
            if k in conditions:
                v = conditions[k]
                # Format info
                # ... special cases
                if k in ['Tvib', 'Tgas', 'Trot']:
                    vs = "{0:.0f}".format(v)
                # ... general case
                elif is_float(v): 
                    vs = "{0:.3g}".format(v)
                else:
                    vs = "{0}".format(v)
                    
                try:
                    un = sjson['cond_units'][k]
                except KeyError: # units not defined, or no units for this condition
                    un = ''
                info.append("{0}{1}{2}".format(k, vs, un))
                # Note: should test for filename validity here.
                # See https://stackoverflow.com/questions/9532499/check-whether-a-path-is-valid-in-python-without-creating-a-file-at-the-paths-ta
                # but it looks long. Best is probably to just test write a file
            else:
                if verbose: print(('Warning. {0} not a valid condition'.format(k)))
        info = '_'.join([_f for _f in info if _f])
    else:
        info = ''
        
    # ... clean from forbidden characters
    for c in [r'/']:
        info = info.replace(c, '')

    # ... get extension
    rad, ext = splitext(name)
    if ext == '':
        ext = '.spec'    # default extension

    # ... Write full name
    name = '_'.join([_f for _f in [date,rad,info] if _f])+ext
    fout = join(fold, name)

    # ... Test for existence, replace if needed
    if exists(fout):
        if if_exists_then=='increment':
            if verbose: print('Warning. File already exists. Filename is incremented')
            i = 0
            while exists(fout):
                i += 1
                name = '_'.join([_f for _f in [date,rad,info,str(i)] if _f])+ext
                fout = join(fold, name)
        elif if_exists_then=='replace':
            if verbose: print(('File exists and will be replaced: {0}'.format(name)))
        else:
            raise ValueError('File already exists {0}. Choose another filename'.format(fout)+\
                             ', or set the `if_exists_then` option to `replace` or ìncrement`')

    return fout

def _compress(s, sjson): 
    ''' if True, removes all quantities that can be regenerated with s.update(),
    e.g, transmittance if abscoeff and path length are given, radiance if
    emisscoeff and abscoeff are given in non-optically thin case, etc.
    Default False '''
    
    
    from radis.spectrum.rescale import get_redundant
    redundant = get_redundant(s)
    
    discarded = []
    for key in list(sjson['q'].keys()):
        if key == 'wavespace': continue
        if redundant[key]:
            del sjson['q'][key]
            discarded.append(key)
    for key in list(sjson['q_conv'].keys()):
        if key == 'wavespace': continue
        if redundant[key]:
            del sjson['q_conv'][key]
            discarded.append(key)
        
    if len(discarded)>0:
        print(('Redundant quantities removed: {0}. Use s.update() after '.format(discarded)+\
              'loading to regenerate them'))
    
    return sjson

def _export_safe(sjson, f, warnings=True):
    ''' Export only jsonable attributes in 'conditions' '''
    
    # remove non jsonable objects
    discard = []
    for k,v in sjson['conditions'].items():
        if not is_jsonable(v):
            discard.append(k)
            if warnings: print(("... discard conditions['{0}'] as non jsonable".format(k)))
    new_conds = {k:v for k,v in sjson['conditions'].items() if not k in discard}
    sjson['conditions'] = new_conds  # dont modify initial spectrum
    
    if 'populations' in list(sjson.keys()):
        # remove non jsonable objects
        # (note: Specair generated populations include numpy arrays tht could
        # be jsonized eventually... just look at what's done with quantities... 
        # but for the time being we just discard them)
        discard = []
        for k,v in sjson['populations'].items():
            if not is_jsonable(v):
                discard.append(k)
                if warnings: print(("... discard populations['{0}'] as non jsonable".format(k)))
        new_pops = {k:v for k,v in sjson['populations'].items() if not k in discard}
        sjson['populations'] = new_pops
    
    # retry export
    json.dump(sjson, f)
    
    
# %% Load functions

def load(file):
    '''
    Parameters
    ----------

    file: str
        .spec file to load

    (wrapper to neq.spec.load)

    '''

    warn(DeprecationWarning("load replaced with more explicit load_spec"))

    return load_spec(file)

def load_spec(file):
    '''
    Parameters
    ----------

    file: str
        .spec file to load

    '''

    with open(file, 'r') as f:
        try:
            sload = json.load(f)
        except:
            print(('Error opening file {0}'.format(f)))
            raise

    # Test format / correct deprecated format:
    sload = _fix_deprecated_format(file, sload)

    # ... Back to real stuff:
    conditions = sload['conditions']
    
    # Get quantities 
    if 'quantities' in sload:
        # old format -saved with tuples (w,I) under 'quantities'): heavier, but 
        # easier to generate a spectrum 
        quantities = {k: (np.array(v[0]), array(v[1]))
                            for (k, v) in sload['quantities'].items()}
        warn("File {0}".format(basename(file))+" has a deprecrated structure ("+\
              "quantities are stored with shared wavespace: uses less space). "+\
            "Regenerate database ASAP.", DeprecationWarning)
    else:
        quantities = {k:(sload['q']['wavespace'],v) for k,v in sload['q'].items() 
                            if k != 'wavespace'}
        quantities.update({k:(sload['q_conv']['wavespace'],v) for k,v in sload['q_conv'].items() 
                            if k != 'wavespace'})

    # Generate spectrum:    
    waveunit = sload['conditions']['waveunit']

    # Only `quantities` and `conditions` is required. The rest is just extra
    # details
    kwargs = {}
    
    # ... load lines if exist
    if 'lines' in sload:
        dtypes = sload.get('_dtypes_lines', True)   # default to True
        if dtypes != True:
            dtypes = pd.read_json(dtypes)
        df = pd.read_json(sload['lines'], dtype=dtypes)
        df.sort_index(inplace=True)
        kwargs['lines'] = df
    else:
        kwargs['lines'] = None
        
    # ... load populations if exist
    if 'populations' in sload:
        pops = {}
        for molecule, isotopes in sload['populations'].items():
            pops[molecule] = {}
            for isotope, states in isotopes.items():
                isotope = int(isotope)    # cast to int
                pops[molecule][isotope] = {}
                for state, content in states.items():
                    pops[molecule][isotope][state] = {}
                    for k, v in content.items():
                        if k in ['vib', 'rovib']:  # pandas dataframes
                            dtypes = content.get('_dtypes_'+k, True)  # default True
                            if dtypes != True:
                                dtypes = pd.read_json(dtypes)
                            df = pd.read_json(v, dtype=dtypes)
                            df.sort_index(inplace=True)
                            pops[molecule][isotope][state][k] = df
                        elif k in ['_dtypes_vib', '_dtypes_rovib']:
                            pass
                        else:
                            pops[molecule][isotope][state][k] = v                        
        kwargs['populations'] = pops
    else:
        kwargs['populations'] = None
        
    # ... load other properties if exist
    for attr in ['units', 'cond_units', 'name']:
        try:
            kwargs[attr] = sload[attr]
        except KeyError:
            kwargs[attr] = None

    return Spectrum(quantities=quantities,
                    conditions=conditions,
                    waveunit=waveunit,
                    **kwargs)

def _fix_deprecated_format(file, sload):
    ''' Test format / correct deprecated format:
    The goal is to still be able to load old format precomputed spectra, and
    fix their attribute names. Save them again to fix the warnigns definitly.
    '''
    try:
        sload['conditions']['waveunit']
    except KeyError as err:
        # deprecation: waveunit was named wavespace
        if 'wavespace' in sload['conditions']:
            sload['conditions']['waveunit'] = sload['conditions']['wavespace']
            del sload['conditions']['wavespace']
        else:
            # temporary fix: on early versions wavespace may be saved as an attribute
            try:
                warn("File {0}".format(basename(file))+" has a deprecrated structure "+\
                      " (attribute 'wgridspace' replaced by 'wavespace' key in 'conditions')."+\
                      "Fixed, but regenerate database ASAP.", DeprecationWarning)
                # Fix it:
                sload['conditions']['wavespace'] = sload['wgridspace']
            except AttributeError:
                raise KeyError("Spectrum 'conditions' dict should at least have a 'waveunit' key")
                
    if 'isotope_identifier' in sload['conditions']:
        warn("File {0}".format(basename(file))+" has a deprecrated structure (key "+\
              "isotope_identifier replaced with isotope). Fixed, but regenerate "+\
            "database ASAP.", DeprecationWarning)
        sload['conditions']['isotope'] = sload['conditions'].pop('isotope_identifier')
        
    if 'air_pressure_mbar' in sload['conditions']:
        warn("File {0}".format(basename(file))+" has a deprecrated structure (key "+\
              "air_pressure_mbar replaced with pressure_mbar). Fixed, but regenerate "+\
            "database ASAP.", DeprecationWarning)
        sload['conditions']['pressure_mbar'] = sload['conditions'].pop('air_pressure_mbar')
        
    if 'isotope' in sload['conditions']:
        isotope = sload['conditions']['isotope']
        if not isinstance(isotope, string_types):
            warn("File {0}".format(basename(file))+" has a deprecrated structure (key "+\
                  "isotope is now a string). Fixed, but regenerate "+\
                "database ASAP.", DeprecationWarning)
            # Fix it:
            sload['conditions']['isotope'] = ','.join([str(k) for k in 
                 list_if_float(isotope)])
    
    if 'dbpath' in sload['conditions']:
        dbpath = sload['conditions']['dbpath']
        if not isinstance(dbpath, string_types):
            warn("File {0}".format(basename(file))+" has a deprecrated structure (key "+\
                  "dbpath is now a string). Fixed, but regenerate "+\
                "database ASAP.", DeprecationWarning)
            # Fix it:
            sload['conditions']['dbpath'] = ','.join([str(k).replace('\\','/') for k in 
                 list_if_float(dbpath)])  # list_if_float or just list??
    
    def fix_path(key):
        if key in sload['conditions']:
            path = sload['conditions'][key]
            if not isinstance(path, string_types):
                warn("File {0}".format(basename(file))+" has a deprecrated structure (key "+\
                      "{0} is now a string). Fixed, but regenerate ".format(key)+\
                    "database ASAP.", DeprecationWarning)
                # Fix it:
                sload['conditions'][key] = path.replace('\\','/')
    fix_path('database')
    fix_path('levelspath')
    fix_path('parfuncpath')
    
    if 'selfabsorption' in sload['conditions']:
        self_absorption = sload['conditions']['selfabsorption']
        sload['conditions']['self_absorption'] = self_absorption
        del sload['conditions']['selfabsorption']
        
    return sload
    

def plot_spec(file, what='radiance', title=True):
    ''' Plot a .spec file. *

    Parameters
    ----------

    file: str
        .spec file to load

    '''

    s = load_spec(file)

    try:
        s.plot(what)
    except KeyError:
        try:
            print((sys.exc_info()[0], sys.exc_info()[1]))
            s.plot(what + '_noslit') # who knows maybe it will work :)
            print(('Printing {0} instead'.format(what + '_noslit')))
        except:
            print((sys.exc_info()[0], sys.exc_info()[1]))
            # Plot something
            s.plot(s.get_vars()[0])
            print(('Printing {0} instead'.format(s.get_vars()[0])))

    if title:
        plt.title(basename(file))
        plt.tight_layout()

    return

# %% Database class
# ... loads database and manipulate it

class SpecDatabase():
    
    def __init__(self, path='.', filt='.spec', add_info=None, add_date='%Y%m%d',
                 verbose=True):
        ''' A Spectrum Database class to manage them all

        It basically manages a list of Spectrum JSON files, adding a Pandas
        dataframe structure on top to serve as an efficient index to visualize
        the spectra input conditions, and slice through the Dataframe with
        easy queries
        
        Parameters
        ----------
        
        path: str
            a folder to initialize the database
            
        filt: str
            only consider files ending with filt
        

        Example
        -------

            >>> db = SpecDatabase(r"path/to/database")     # create or loads database

            >>> db.update()  # in case something changed
            >>> db.see(['Tvib', 'Trot'])   # nice print in console

            >>> s = db.get('Tvib==3000')[0]  # get a Spectrum back
            >>> db.add(s)  # update database (and raise error because duplicate!)

        Note that SpectrumFactory objects can be configured to automatically update
        a database. 
        
        The function to auto retrieve a Spectrum from database on calculation
        time is a method of DatabankLoader class
        
        Note
        ----
        
        Database interaction is based on Pandas query functions. As such, it 
        requires all conditions to be either float, string, or boolean. List 
        won't work!

        '''

        # Assert name looks like a directory
        name, ext = splitext(path)

        if ext != '':
            raise ValueError('Database should be a directory: {0}'.format(path))

        if not exists(path):
            # create it
            os.mkdir(path)
            if verbose: print(('Database {0} initialised in {1}'.format(basename(path),
                              dirname(path))))
        else:
            if verbose: print(('Loading database {0}'.format(basename(path))))

        self.name = basename(name)
        self.path = path
        self.df = None
        self.verbose = verbose

        # default
        self.add_info = add_info
        self.add_date = add_date

        self.update(force_reload=True, filt=filt)

    def conditions(self):
        ''' Show conditions in database '''

        cond = list(self.df.columns)

        if len(cond) > 0:    
            cond.remove('file')
            cond.remove('Spectrum')

        return cond

    def see(self, columns=None, *args):
        ''' Shows Spectrum database will whole conditions (index=None) or
        specific conditions

        Parameters
        ----------

        columns: str, list of str, or None
            shows the conditions value for all cases in database. If None, all
            conditions are shown. Default None
            e.g.
            >>> db.see(['Tvib', 'Trot'])

        Notes
        -----

        Makes the 'file' column the index, and also discard the 'Spectrum' column
        (that holds all the data) for readibility

        '''

        if len(self) == 0:
            raise ValueError('Database is empty')

        if type(columns)==str:
            columns = [columns]+[k for k in args]
            
        dg  = self.df.set_index('file')  # note that this is a copy already.
                                    # dont try to modify the output of "see"
        del dg['Spectrum']          # for visibility"

        if columns is None: 
            return dg
        
        for c in columns:
            if not c in self.df.columns:
                raise ValueError('`{0}` is not a column name. Use one of {1}'.format(c, 
                                 self.df.columns))
            
        return dg.reindex(columns=columns)

    def view(self, columns=None, *args):
        ''' alias to df.see() '''
        
        return self.see(columns = columns, *args)

    def update(self, force_reload=False, filt='.spec'):
        ''' Reloads database, updates internal index structure and print it
        in <database>.csv

        Parameters
        ----------

        force_reload: boolean
            if True, reloads files already in database

        filt: str
            only consider files ending with `filt`
            
        '''

        path = self.path

        if force_reload:
            # Reloads whole database  (necessary on database init to create self.df
            files = [join(path, f) for f in os.listdir(path) if f.endswith(filt)]
            self.df = self._load_files(files=files)
        else:
            dbfiles = list(self.df['file'])
            files = [join(path, f) for f in os.listdir(path) if f not in dbfiles
                     and f.endswith(filt)]
            for f in files:
                self.df = self.df.append(self._load_file(f), ignore_index=True)

        # Print index
        self.print_index()

    def print_index(self, file=None):
        if file is None:
            file = join(self.path, basename(self.path)+'.csv')

        if len(self) > 0:
            try:
                self.see().to_csv(file)
            except PermissionError:
                warn('Database index could not be updated: {0}'.format(sys.exc_info()[1]))
        else:
            try:
                os.remove(file)  # if database existed but files were deleted
            except PermissionError:
                warn('Database index could not be updated: {0}'.format(sys.exc_info()[1]))
            except FileNotFoundError:
                pass

    def find_duplicates(self, columns=None):
        ''' Find spectra with same conditions '''

        dg = self.see(columns=columns).duplicated()


        if columns is None:
            columns = 'all'

        if self.verbose: print(('{0} duplicate(s) found'.format(dg.sum())+\
                               ' based on columns: {0}'.format(columns)))
        return dg

    def add(self, spectrum, **kwargs):
        ''' Add Spectrum to database, whether it's a Spectrum object or a file
        that stores one. Check it's not in database already.

        Parameters
        ----------

        spectrum: path to .spec file, or Spectrum object
            if Spectrum object: stores it in database first (using Spectrum.store()),
            then adds the file
            if path to file: will first copy the file in database folder, then
            adds the file

        **kwargs: **dict
            extra parameters used in the case where spectrum is a file and a .spec object
            has to be created (useless if `spectrum` is a file already). kwargs
            are forwarded to Spectrum.store() class. See Spectrum.store() or
            database.save() for more information

        '''

        # Get file in database
        # ... input is a Spectrum. Store it in database and load it from there
        if type(spectrum) is Spectrum:
            # add defaults
            if not 'add_info' in kwargs:
                kwargs['add_info'] = self.add_info
            if not 'add_date' in kwargs:
                kwargs['add_date'] = self.add_date
            file = spectrum.store(self.path, **kwargs)
            # Note we could have added the Spectrum directly
            # (saves the load stage) but it also serves to
            # check the file we just stored is readable

        # ... input is a file. Copy it in database and load it
        elif type(spectrum) is str:
            if not exists(spectrum):
                raise FileNotFoundError('File doesnt exist: {0}'.format(spectrum))

            fd, name = split(spectrum)

            # Assert a similar case name is not in database already
            if spectrum in list(self.df['file']):
                raise ValueError('File already in database: {0}. Database filenames should be unique'.format(
                        spectrum))

            if abspath(fd) != abspath(self.path):
                # Assert file doesnt exist in database already
                if name in os.listdir(self.path):
                    raise ValueError('File already in database folder: {0}'.format(name)+\
                                     '. Use db.update() if you added it there manually')
                # Ok. Copy it.
                file = join(self.path, name)
                copy2(spectrum, file)

            else:
                warn('You are manually adding a file that is in the database folder directly. Consider using db.update()')
                file = spectrum

        else:
            raise ValueError('Unvalid Spectrum type: {0}'.format(type(spectrum)))

        # Add file
        self.df = self.df.append(self._load_file(file), ignore_index=True)

        # Update index .csv
        self.print_index()

        return

    def _load_files(self, files):
        ''' Parse files and generate a database
        '''
        db = []
        for f in files:
            db.append(self._load_file(f))

        return pd.DataFrame(db)

    def _load_file(self, file):
        ''' return Spectrum attributes for insertion in database.
        '''

        s = load_spec(file)

        if self.verbose: print(('loaded {0}'.format(basename(file))))

        out = s.get_conditions().copy()
        
        # Add filename, and a link Spectrum itself
        out.update({'file':basename(file), 'Spectrum':s})

        return out


    def get(self, conditions='', **kwconditions):
        '''   Returns a list of spectra that match given conditions

        Parameters
        ----------

        database: list of Spectrum objects
            the database

        conditions: str
            a list of conditions
            >>> get('Tvib==3000 & Trot==1500')

        kwconditions: dict
            an unfolded dict of conditions
            >>> get(Tvib=3000, Trot=1500)
            
        Other Parameters
        ----------------
        
        inplace: boolean
            if True, return the actual object in the database. Else, return
            copies. Default False

        Examples
        --------

        >>> spec_list = db.get('Tvib==3000 & Trot==1300')
        
        or
        
        >>> spec_list = db.get(Tvib=3000, Trot=1300)

        '''
                
        # Test inputs
        for (k,_) in kwconditions.items():
            if not k in self.df.columns:
                raise ValueError('{0} not a correct condition name. Use one of: {1}'.format(k, 
                                 self.df.columns))
        if len(self.df) == 0:
            warn('Empty database')
            return []
        
        inplace = kwconditions.pop('inplace', False)         # type: bool, default False

        # Unique condition method
        if conditions != '' and kwconditions != {}:
            raise ValueError("Please choose one of the two input format (str or dict) exclusively")

        if conditions == '' and kwconditions == {}:
            return list(self.df['Spectrum'])

        if conditions != '':
            dg = self.df.query(conditions)
        else:
            query = []
            for (k, v) in kwconditions.items():
                if isinstance(v, string_types):
                    query.append("{0} == '{1}'".format(k, v))
                else:
                    query.append('{0} == {1}'.format(k,v))
            # There is a limitation in numpy: a max of 32 arguments is required. 
            # Below we write a workaround when the Spectrum has more than 32 conditions
            if len(query) < 32:
                query = ' & '.join(query)
                if __debug__: printdbg('Database query: {0}'.format(query))
                dg = self.df.query(query)
            else:
                # cut in <32-long parts
                N = len(query)//32+1
                querypart = ' & '.join(query[::N])
                dg = self.df.query(querypart)
                for i in range(1, N+1):
                    querypart = ' & '.join(query[i::N])
                    if __debug__: printdbg('Database query: {0}'.format(querypart))
                    dg = dg.query(querypart)
            
        out = list(dg['Spectrum'])
                    
        if not inplace:
            out = [s.copy() for s in out]
        
        return out

    def get_unique(self, conditions='', **kwconditions):
        ''' Returns a spectrum that match given conditions. Raises an error
        if the spectrum is not unique 
        
        Parameters
        ----------
        
        see .get() for more details '''
        
        out = self.get(conditions, **kwconditions)

        if len(out) == 0:
            raise ValueError('Spectrum not found')
        elif len(out) > 1:
            raise ValueError('Spectrum is not unique ({0} match found)'.format(
                    len(out)))
        else:
            return out[0]

    def get_closest(self, scale_if_possible=True, **kwconditions):
        '''   Returns the Spectra in the database that is the closest to the input conditions

        Note that for non-numeric values only equals should be given.
        To calculate the distance all numeric values are scaled by their
        mean value in the database
        
        Parameters
        ----------
        
        kwconditions: named arguments 
            i.e: Tgas=300, path_length=1.5
            
        scale_if_possible: boolean
            if True, spectrum is scaled for parameters that can be computed 
            directly from spectroscopic quantities (e.g: 'path_length', 'molar_fraction')
            Default True
            
        Extra Parameters
        ----------------
        
        verbose: boolean
            print messages. Default True
            
        inplace: boolean
            if True, returns the actual object in database. Else, return a copy
            Default False
            
        '''
        
        scalable_inputs = ['mole_fraction', 'path_length']

        # Test inputs
        if kwconditions == {}:
            raise ValueError('Please specify filtering conditions. e.g: Tgas=300')
        
        # Inputs:
        verbose = kwconditions.pop('verbose', True)         # type: bool
        inplace = kwconditions.pop('inplace', False)        # type: bool
        
        for (k,_) in kwconditions.items():
            if not k in self.df.columns:
                raise ValueError('{0} not a correct condition name. Use one of: {1}'.format(k, 
                                 self.df.columns))

        dg = self.df.reindex(columns=list(kwconditions))
        
        if scale_if_possible:
            # Remove scalable inputs from distance calculation variables (unless
            # they're the last variable, because then it may screw things up)
            for k in scalable_inputs:
                if len(dg.columns) == 1:
                    break
                try:
                    del dg[k]
                except KeyError:
                    pass
                
#        raise
        
        mean = dict(dg.mean())
        std = dict(dg.std())
        dg['_d'] = 0  # distance (squared, actually)
        for k, v in kwconditions.items():
            if not k in dg.columns:
                continue
#            # add distance to all conditions planes. We regularize the different
#            # dimensions by working on normalized quantities: 
#            #     a  ->   (a-mean)/std  € [0-1]
#            # Distance becomes:
#            #     d^2 ->  sum((a-target)/std)^2
#            # Problem when std == 0! That means this dimension is not discrimant
#            # anyway
#            if std[k] == 0:
#                # for this conditions all parameters have the same value. 
#                dg['_d'] += (dg[k]-v)**2
#            else:
#                dg['_d'] += (dg[k]-v)**2/(std[k])**2
            # Eventually I chose to scale database with mean only (there was
            # an obvious problem with standard deviation scaling in the case of 
            # a non important feature containing very close datapoints that would
            # result in inappropriately high weights)
            dg['_d'] += (dg[k]-v)**2/mean[k]**2

        # self.plot('Tvib', 'Trot', dg['_d'])  # for debugging
        
        # Get spectrum with minimum distance to target conditions
        sout = self.df.ix[dg['_d'].idxmin(), 'Spectrum']     # type: Spectrum
        
        if not inplace:
            sout = sout.copy()

        # Scale scalable conditions
        if scale_if_possible:
            if 'path_length' in kwconditions:
                sout.rescale_path_length(kwconditions['path_length'])
            if 'mole_fraction' in kwconditions:
                sout.rescale_mole_fraction(kwconditions['mole_fraction'])

        if verbose:
            print(('------- \t'+'\t'.join(['{0}'.format(k) for k in kwconditions.keys()])))
            print(('Look up \t'+'\t'.join(['{0:.3g}'.format(v) for v in kwconditions.values()])))
            print(('Got     \t'+'\t'.join(['{0:.3g}'.format(sout.conditions[k]) for k in kwconditions.keys()])))

        return sout

    def plot(self, cond_x, cond_y, z_value=None, nfig=None):
        ''' Plot database points available:

        Parameters
        ----------

        cond_x, cond_y: str
            columns (conditions) of database.

        z_value: array, or None
            if not None, colors the 2D map with z_value. z_value is ordered
            so that z_value[i] corresponds to row[i] in database.

        Examples
        --------
    
        >>> db.plot(Tvib, Trot)     # plot all points calculated
    
    
        >>> db.plot(Tvib, Trot, residual)     # where residual is calculated by a fitting
                                              # procedure...

        '''
        # %%

        x = self.df[cond_x]
        y = self.df[cond_y]

        # Default
        fig = plt.figure(num=nfig)
        ax = fig.gca()
        ax.plot(x, y, 'ok')
        ax.set_xlabel(cond_x)
        ax.set_ylabel(cond_y)
        title = basename(self.path)

        # Overlay color
        if z_value is not None:
            assert(len(z_value))==len(self.df)

            z = np.array(z_value)**0.5  # because the lower the better

#            norm = cm.colors.Normalize(vmax=z.max(), vmin=z.min())
#            cmap = cm.PRGn


            xarr = np.linspace(min(x),max(x))
            yarr = np.linspace(min(y),max(y))
            mx, my = np.meshgrid(xarr,yarr)
            zgrid = griddata((x,y),z,(mx,my),method='linear')
            levels=np.linspace(min(z), max(z), 20)
            cs0 = ax.contourf(mx,my,zgrid,  levels=levels,
                               #linewidths=1,linestyles='dashed',
                               extend='both')

            ix0, iy0 = np.where(zgrid == zgrid.min())
            plt.plot((xarr[ix0], xarr[ix0]), (yarr.min(), yarr.max()), color='white', lw=0.5)
            plt.plot((xarr.min(), xarr.max()), (yarr[iy0], yarr[iy0]), color='white', lw=0.5)
            # lbls = plt.clabel(cs0, inline=1, fontsize=16,fmt='%.0fK',colors='k',manual=False)     # show labels

                              # %%

        plt.title(title)
        plt.tight_layout()

    def __len__(self):
        return len(self.df)



def in_database(smatch, db='.', filt='.spec'):
    ''' old function '''
    match_cond = smatch.get_conditions()
    for f in [f for f in os.listdir(db) if f.endswith(filt)]:
        fname = join(db, f)
        s = load_spec(fname)
        if s.get_conditions() == match_cond:
            return True
    return False

# %% Test
if __name__ == '__main__':

    from neq.test.spec.test_database import _run_testcases
    print(('Testing database.py: ', _run_testcases()))


# -*- coding: utf-8 -*-
"""
Created on Thu Sep 21 19:10:01 2017

@author: erwan

Classes to deal with multi-slabs configurations: 
- MergeSlabs for several Spectrum
- SerialSlabs to add several spectra along the line-of-path

One Slab is just a Spectrum object 

Todo
------

- transport emisscoeff too 
- emisscoeff default unit 

"""

from __future__ import print_function, absolute_import, division, unicode_literals

from radis.spectrum.spectrum import Spectrum, is_spectrum, NON_CONVOLUTED_QUANTITIES
from radis.phys.convert import nm2cm
from radis.misc.basics import merge_lists
from radis.misc.utils import FileNotFoundError
from radis.misc.arrays import count_nans
from radis.misc.debug import printdbg
from warnings import warn
import numpy as np
from numpy import exp, arange, allclose, abs, diff

# %% Slabs / Multi-layers / Radiative Transfer Equation (RTE)
#----------------------------------------------------------------------

def intersect(a, b):
    ''' Returns intersection of two dictionaries on values'''
    c = {}
    for k in set(a.keys()) & set(b.keys()): # work in Python 2?
        c[k] = a[k] if (a[k] == b[k]) else 'N/A'
    return c

def SerialSlabs(*slabs, **kwargs):
    ''' Compute the result of several slabs 

    
    Parameters    
    ----------
    
    slabs: list of Spectra, each representing a slab
        slabs       [0]     [1]  ............... [n]     
                     :       :                    :         \====
        light        *   ->  *        ->          *    ->    )===  observer
                                                            /====
            
    resample_wavespace: 'never', 'intersect', 'full'
        what to do when spectra have different wavespaces. 
        - If 'never', raises an error
        - If 'intersect', uses the intersection of all ranges, and resample
        spectra on the most resolved wavespace. 
        - If 'full', uses the overlap of all ranges, resample spectra on the 
        most resolved wavespace, and fill missing data with 0 emission and 0
        absorption
        Default 'never'
            
    out_of_bounds: 'transparent', 'nan', 'error'
        what to do if resampling is out of bounds. 'transparent': fills with 
        transparent medium. 'nan': fills with nan. 'error': raises an error. 
        Default 'nan'
                       

    Returns
    -------
    
    Spectrum object representing total emission and total transmittance as 
    observed at the output (slab[n+1]). Conditions and units are transported too,
    unless there is a mismatch then conditions are dropped (and units mismatch
    raises an error because it doesnt make sense)
    
    Use
    -------
    
        >>> s1 = calc_spectrum(...)
        >>> s2 = calc_spectrum(...)
        >>> s3 = SerialSlabs(s1, s2)
        
    Todo
    -------
    
    rewrite with 'recompute' list like in MergeSlabs
    
    '''

    # Check inputs, get defaults
    resample_wavespace = kwargs.get('resample_wavespace', 'never')   # default 'never'
    out_of_bounds = kwargs.get('out_of_bounds', 'nan')               # default 'nan'
    
    if resample_wavespace not in ['never', 'intersect', 'full']:
        raise ValueError("resample_wavespace should be one of: {0}".format
                         (', '.join(['never', 'intersect', 'full'])))
        
    if len(slabs)==0:
        raise ValueError('Empty list of slabs')
    
    elif len(slabs)==1:
        if not is_spectrum(slabs[0]):
            raise TypeError('SerialSlabs takes an unfolded list of Spectrum as '+\
                                 'argument: *list (got {0})'.format(type(slabs[0])))
        return slabs[0]

    else:  
        # recursively calculate serial slabs
        slabs = list(slabs)
        
        # TODO: correct class but it crashes anyway. 
        #  "All inputs must be Spectrum objects (got: <class 'neq.spec.spectrum.Spectrum'>)"
        
#        # Check all items are Spectrum
        for s in slabs:
            _check_valid(s)
            
        sn = slabs.pop(-1)            # type: Spectrum
        s = SerialSlabs(*slabs, resample_wavespace=resample_wavespace,
                        out_of_bounds=out_of_bounds)
        
        quantities = {}
        
        # make sure we use the same wavespace type (even if sn is in 'nm' and s in 'cm-1')
        waveunit = s.get_waveunit()  
        w = s.get('radiance_noslit', wunit=waveunit)[0]
        wn = sn.get('radiance_noslit', wunit=waveunit)[0]
        
        # Make all our slabs copies with the same wavespace range
        # (note: wavespace range may be different for different quantities, but
        # equal for all slabs)
        s, sn = resample_slabs(waveunit, resample_wavespace, out_of_bounds, 
                               s, sn)
        
        w, I = s.get('radiance_noslit', wunit=waveunit)  
        wn, In = sn.get('radiance_noslit', wunit=waveunit)
        _, Tn = sn.get('transmittance_noslit', wunit=waveunit)
        
        units = intersect(s.units, sn.units)
        
        if 'radiance_noslit' in s._q and 'radiance_noslit' in sn._q:
            # case where we may use SerialSlabs just to compute the products of all transmittances
            # Check units:
            unit = s.units['radiance_noslit']
            unitn = sn.units['radiance_noslit']
            if (unit != unitn):
                # convert unit of (w,I) into (wn, In)
                I = s.get_radiance_noslit(unitn)  # will crash if not possible
#                I = conv2(I, unit, unitn)  # will crash if not possible
                units['radiance_noslit'] = unitn
            quantities['radiance_noslit']=(w, I * Tn + In)
            
        if 'transmittance_noslit' in s._q:  # note that we dont need the transmittance in the inner
                                         # slabs to calculate the total radiance
            _, T = s.get('transmittance_noslit', wunit=waveunit)
            # Check units:
            if (s.units['transmittance_noslit'] != sn.units['transmittance_noslit']):
                # actually transmittance should have no units always...
                raise ValueError('Transmittance units arent the same in SerialSlabs '+\
                                 'calculation: {0} & {1}'.format(
                        s.units['radiance_noslit'], sn.units['radiance_noslit'])) 
            quantities['transmittance_noslit']=(w, Tn * T)
        
        # Get conditions (if they're different, fill with 'N/A')
        conditions = intersect(s.conditions, sn.conditions)
        conditions['waveunit'] = waveunit
        
        cond_units = intersect(s.cond_units, sn.cond_units)
        
        # name
        name = _serial_slab_names(s,sn)
        
        return Spectrum(quantities=quantities, conditions = conditions, 
                        cond_units = cond_units, units = units,
                        name=name)

def _serial_slab_names(s, sn):
    name_s = s.get_name()
    name_sn = sn.get_name()
    if '//' in name_s and not '>>' in name_s:
        name_s = '({0})'.format(name_s)
    if '//' in name_sn:
        name_sn = '({0})'.format(name_sn)
    return '{0}>>{1}'.format(name_s, name_sn)

def _check_valid(s):
    ''' Check s is a valid Spectrum object. Raises an error if not '''
    
    if not is_spectrum(s):
        raise TypeError('All inputs must be Spectrum objects (got: {0})'.format(
                type(s)))
    sdict = s._get_items()
    for k in sdict.keys():
        if count_nans(sdict.get(k))>0:
            warn('Nans detected in Spectrum object for multi-slab operation. '+\
                 'Results may be wrong!')
    
    return True

def _has_quantity(quantity, *slabs):
    slabs = list(slabs)
    b = True
    for s in slabs:
        b *= quantity in slabs
    return b

def resample_slabs(waveunit, resample_wavespace, out_of_bounds='nan', *slabs):
    ''' Resample slabs on the same wavespace: if the range are differents, 
    depending on the mode we may fill with optically thin media, or raise an
    error 
    
    
    Parameters    
    ----------
    
    waveunit: 'nm', 'cm-1'
        which wavespace we're working in 
    
    resample_wavespace: 'never', 'intersect', 'full'
        what to do when spectra have different wavespaces. 
        - If 'never', raises an error
        - If 'intersect', uses the intersection of all ranges, and resample
        spectra on the most resolved wavespace. 
        - If 'full', uses the overlap of all ranges, resample spectra on the 
        most resolved wavespace, and fill missing data with 0 emission and 0
        absorption
        Default 'never'
        
    out_of_bounds: 'transparent', 'nan', 'error'
        what to do if resampling is out of bounds. 'transparent': fills with 
        transparent medium. 'nan': fills with nan. 'error': raises an error. 
        Default 'nan'
        
    *slabs: list of Spectrum objects
    

    Returns
    -------
    
    slabs: list of Spectrum objects
        resampled copies of inputs Spectra. All now have the same wavespace
        
    ''' 

    # Check wavespace is the same
    def same_wavespace(wl):
        if not all([len(w) == len(wl[0]) for w in wl[1:]]):
            return False
        elif not all([allclose(w, wl[0]) for w in wl[1:]]):
            return False
        else:
            return True
        
    # Work on copies
    slabs = [s.copy() for s in slabs]
    
    # Get all keys
    keys = merge_lists([s.get_vars() for s in slabs])
    
    for k in keys:
        # We ensure all quantities are resampled. Note that .resample() deals 
        # with all quantities so usually spectra are fully corrected after the 
        # first iteration. However, it also deals with cases where quantities 
        # have different wavespaces (ex: radiance and radiance_noslit are not
        # defined on the same range)
        slabsk = [s for s in slabs if k in s.get_vars()]  # note that these are
                                        # references to the actual Spectrum copy
        wl = [s.get(k, wunit=waveunit)[0] for s in slabsk]
        if not same_wavespace(wl):
            # resample slabs if allowed
            if resample_wavespace == 'never':
                raise ValueError('All wavelengths/wavenumbers must be the same for '+\
                                 'multi slabs configurations. Consider using `resample_wavespace` '+\
                                 '= `intersect` or `full`')
            elif resample_wavespace == 'full':
                # ... get bounds
                wmin = min([w.min() for w in wl])   # minimum of all 
                wmax = max([w.max() for w in wl])   # maximum of all 
                dw = min([abs(diff(w)).min() for w in wl])  # highest density
                wnew = arange(wmin, wmax+dw, dw)
                # ... copy the array of slabs not to modify the input
                # ... slabs themselves
                for s in slabsk:
                    s.resample(wnew, unit=waveunit, if_conflict_drop='convoluted',
                               out_of_bounds=out_of_bounds)
                # note: s.resample() fills with 0 when out of bounds
            elif resample_wavespace == 'intersect':
                # ... get bounds
                wmin = max([w.min() for w in wl])   # maximum of all 
                wmax = min([w.max() for w in wl])   # minimum of all 
                dw = min([abs(diff(w)).min() for w in wl])  # highest density
                wnew = arange(wmin, wmax+dw, dw)
                if len(wnew)==0:
                    raise ValueError('Intersect range is empty')
                # ... copy the array of slabs not to modify the input
                # ... slabs themselves
                for s in slabsk:
                    s.resample(wnew, unit=waveunit, if_conflict_drop='convoluted',
                               out_of_bounds=out_of_bounds)
                # note: s.resample() fills with 0 when out of bounds
        # Now all our slabs have the same wavespace
            
    return slabs
    
def MergeSlabs(*slabs, **kwargs):
    ''' Combines several slabs into one. Useful to calculate multi-gas slabs. 
    Linear absorption coefficient is calculated as the sum of all linear absorption
    coefficients, and the RTE is recalculated to get the total radiance

    
    Parameters    
    ----------
    
    slabs: list of Spectra, each representing a slab
        If given in conditions, all path_length have to be same

    kwargs input
    --------

    resample_wavespace: 'never', 'intersect', 'full'
        what to do when spectra have different wavespaces. 
        - If 'never', raises an error
        - If 'intersect', uses the intersection of all ranges, and resample
        spectra on the most resolved wavespace. 
        - If 'full', uses the overlap of all ranges, resample spectra on the 
        most resolved wavespace, and fill missing data with 0 emission and 0
        absorption
        Default 'never'
        
    out_of_bounds: 'transparent', 'nan', 'error'
        what to do if resampling is out of bounds. 'transparent': fills with 
        transparent medium. 'nan': fills with nan. 'error': raises an error. 
        Default 'nan'

    optically_thin: boolean
        if True, merge slabs in optically thin mode. Default False 
                            
    verbose: boolean
        if True, print messages and warnings. Default True
        

    Returns
    -------
    
    Spectrum object representing total emission and total transmittance as 
    observed at the output. Conditions and units are transported too,
    unless there is a mismatch then conditions are dropped (and units mismatch
    raises an error because it doesnt make sense)
    
    
    Examples
    -------
    
    Merge two spectra calculated with different species (true only if broadening
    coefficient dont change much):
    
        >>> from radis import calc_spectrum, MergeSlabs
        >>> s1 = calc_spectrum(...)
        >>> s2 = calc_spectrum(...)
        >>> s3 = MergeSlabs(s1, s2)
        
    Load a spectrum precalculated on several partial spectral ranges, for a same 
    molecule (i.e, partial spectra are optically thin on the rest of the spectral 
    range)
    
        >>> from radis import load_spec, MergeSlabs
        >>> spectra = []
        >>> for f in ['spec1.spec', 'spec2.spec', ...]:
        >>>     spectra.append(load_spec(f))
        >>> s = MergeSlabs(*spectra, resample_wavespace='full', out_of_bounds='transparent')
        >>> s.update()   # Generate missing spectral quantities
        >>> s.plot()
        
    '''
    
    # inputs (Python 2 compatible)
    resample_wavespace = kwargs.get('resample_wavespace', 'never')   # default 'never'
    out_of_bounds = kwargs.get('out_of_bounds', 'nan')               # default 'nan'
    optically_thin = kwargs.pop('optically_thin', False)             # default False
    
    verbose = kwargs.pop('verbose', True)             # type: bool
    debug = kwargs.pop('debug', False)                # type: bool
    
    # Check inputs
    if resample_wavespace not in ['never', 'intersect', 'full']:
        raise ValueError("resample_wavespace should be one of: {0}".format
                         (', '.join(['never', 'intersect', 'full'])))
        
    if len(slabs)==0:
        raise ValueError('Empty list of slabs')
    
    elif len(slabs)==1:
        if not is_spectrum(slabs[0]):
            raise TypeError('MergeSlabs takes an unfolded list of Spectrum as '+\
                                 'argument: (got {0})'.format(type(slabs[0])))
        return slabs[0]

    else:  # calculate serial slabs
        
        slabs = list(slabs)
        
#        # Check all items are valid Spectrum objects
        for s in slabs:
            _check_valid(s)
        
        # Just check all path_lengths are the same if they exist
        try:
            path_lengths = [s.conditions['path_length'] for s in slabs]
        except KeyError:
            raise KeyError('path_length must be defined for all MergeSlabs inputs')
        if not all([L == path_lengths[0] for L in path_lengths[1:]]):
            raise ValueError('path_length must be equal for all MergeSlabs inputs'+\
                             '  (got {0})'.format(path_lengths))
        
        # make sure we use the same wavespace type (even if sn is in 'nm' and s in 'cm-1')
        waveunit = slabs[0].get_waveunit()  
        # Make all our slabs copies with the same wavespace range
        # (note: wavespace range may be different for different quantities, but
        # equal for all slabs)
        slabs = resample_slabs(waveunit, resample_wavespace, out_of_bounds, *slabs)
        w_noconv = slabs[0]._get_wavespace()
    
        # %%
        for s in slabs:
            if not 'abscoeff' in list(s._q.keys()):
                try:
                    s.get('absorbance', wunit=waveunit)
                    s.conditions['path_length']
#                    s['abscoeff'] = (w, A/s.conditions['path_length'])
                except KeyError:
                    raise KeyError('If abscoeff is not a calculated spectrum quantity '+\
                                   'then both `absorbance` and conditions[`path_length`] '+\
                                   'should be given for MergeSlabs inputs')
        
        
        # Get conditions
        conditions = slabs[0].conditions
        conditions['waveunit'] = waveunit
        cond_units = slabs[0].cond_units
        units = slabs[0].units
        for s in slabs[1:]:
            conditions = intersect(conditions, s.conditions)
            cond_units = intersect(cond_units, s.cond_units)
            units = intersect(units, s.units)
        
        # %% Get quantities that should be calculated 
            
        requested = merge_lists([s.get_vars() for s in slabs])
        recompute = requested.copy()
        if ('radiance_noslit' in requested and not optically_thin):
            recompute.append('emisscoeff')
            recompute.append('abscoeff')
        if 'abscoeff' in recompute and 'path_length' in conditions:
            recompute.append('absorbance')
            recompute.append('transmittance_noslit')
                    
        path_length = conditions['path_length']

        # %% Calculate new quantites
        added = {}        
        
        # ... absorption coefficient (cm-1)
        if 'abscoeff' in recompute:
            #TODO: deal with all cases
            if __debug__: printdbg('... merge: calculating abscoeff k=sum(k_i)')
            abscoeff_eq = np.sum([s.get('abscoeff', wunit=waveunit)[1] for s in slabs], axis=0)
            assert len(w_noconv) == len(abscoeff_eq)
            added['abscoeff'] = (w_noconv, abscoeff_eq)
        
        if 'absorbance' in recompute:
            if 'abscoeff' in added:
                if __debug__: printdbg('... merge: calculating absorbance A=k*L')
                _, abscoeff_eq = added['abscoeff']            
                absorbance_eq = abscoeff_eq*path_length
            else:
                raise NotImplementedError('recalculate abscoeff first')
            added['absorbance'] = (w_noconv, absorbance_eq)
        
        # ... transmittance
        if 'transmittance_noslit' in recompute:    
            if 'absorbance' in added:
                if __debug__: printdbg('... merge: calculating transmittance T=exp(-A)')
                _, absorbance_eq = added['absorbance']
                transmittance_noslit_eq = exp(-absorbance_eq)
            else:
                raise NotImplementedError('recalculate absorbance first')
            added['transmittance_noslit'] = (w_noconv, transmittance_noslit_eq)

        # ... emission coefficient
        if 'emisscoeff' in recompute:
            emisscoeff_eq = np.zeros_like(w_noconv)
            for i, s in enumerate(slabs):
                # Manual loop in case all Slabs dont have the same keys
                # Could also do a slab.update() first then sum emisscoeff directly
                if 'emisscoeff' in list(s._q.keys()):
                    if __debug__: printdbg('... merge: calculating emisscoeff: j+=j_i')
                    _, emisscoeff = s.get('emisscoeff', wunit=waveunit)
                elif optically_thin and 'radiance_noslit' in list(s._q.keys()):
                    if __debug__: printdbg('... merge: calculating emisscoeff: j+=I_i/L '+\
                                    '(optically thin case)')
                    _, I = s.get('radiance_noslit', wunit=waveunit)
                    emisscoeff = I/path_length
                    emisscoeff_eq += emisscoeff
                else:
                    wI, I = s.get('radiance_noslit', wunit=waveunit)
                    if __debug__: printdbg('... merge: calculating emisscoeff j+=[k*I/(1-T)]_i)')
                    try:
                        wT, T = s.get('transmittance_noslit', wunit=waveunit)
                        wk, k = s.get('abscoeff', wunit=waveunit)
                    except KeyError:
                        raise KeyError('Need transmittance_noslit and abscoeff to '+\
                                       'recompute emission coefficient')
                    b = (T == 1)  # optically thin mask
                    emisscoeff = np.zeros_like(T)
                    emisscoeff[b] = I[b]/path_length     # optically thin case
                    emisscoeff[~b] = I[~b]/(1-T[~b])*k[~b]
                emisscoeff_eq += emisscoeff
            added['emisscoeff'] = (w_noconv, emisscoeff_eq)
            
        # ... derive global radiance (result of analytical RTE solving)
        if 'radiance_noslit' in recompute:
            if 'emisscoeff' in added and optically_thin:
                if __debug__: printdbg('... merge: recalculating radiance_noslit I=j*L(optically_thin)')
                (_, emisscoeff_eq) = added['emisscoeff']
                radiance_noslit_eq = emisscoeff_eq * path_length  # optically thin 
            elif ('emisscoeff' in added and 'transmittance_noslit' in added and
                'abscoeff' in added):
                if __debug__: printdbg('... merge: recalculating radiance_noslit I=j/k*(1-T)')
                (_, emisscoeff_eq) = added['emisscoeff']
                (_, abscoeff_eq) = added['abscoeff']
                (_, transmittance_noslit_eq) = added['transmittance_noslit']
                b = (abscoeff_eq == 0) # optically thin mask
                radiance_noslit_eq = np.zeros_like(emisscoeff_eq)
                radiance_noslit_eq[b] = emisscoeff_eq[b] * path_length  # optically thin limit
                radiance_noslit_eq[~b] = emisscoeff_eq[~b]/abscoeff_eq[~b]*(1-
                                  transmittance_noslit_eq[~b])
            elif optically_thin:
                if __debug__: printdbg('... merge: recalculating radiance_noslit I=sum(I_i) (optically thin)')
                radiance_noslit_eq = np.zeros_like(w_noconv)
                for s in slabs:
                    if 'radiance_noslit' in list(s._q.keys()):
                        radiance_noslit_eq += s.get('radiance_noslit', wunit=waveunit)[1]
                    else:
                        raise KeyError('Need radiance_noslit for all slabs to '+\
                                       'recalculate for the MergeSlab (could also '+\
                                       'get it from emisscoeff but not implemented)')
            else:
                if optically_thin:
                    raise ValueError('Missing data to recalculate radiance_noslit'+\
                                     '. Try optically_thin mode?')
                else:
                    raise ValueError('Missing data to recalculate radiance_noslit')
            added['radiance_noslit'] = (w_noconv, radiance_noslit_eq)
        
        # ... emissivity no slit
        if 'emissivity_noslit' in requested:
            added['emissivity_noslit'] = w_noconv, np.ones_like(w_noconv)*np.nan
#            if verbose: 
#                warn('emissivity dropped during MergeSlabs') 
            # Todo: deal with equilibrium cases?
        
        # Store output
        quantities = {}
        for k in requested:
            quantities[k] = added[k]
            
        # name
        name = '//'.join([s.get_name() for s in slabs])
        
        # TODO: check units are consistent in all slabs inputs
        return Spectrum(quantities=quantities, conditions = conditions, 
                        cond_units = cond_units, units = units,
                        name=name)

# %% Tests
        
def _test_merge_slabs(verbose=True, plot=True, warnings=True, debug=False, 
                      *args, **kwargs):
    ''' Merge 10 slabs with 1/10 of concentration, and compare the results. 
    Ensure error is < 0.1%
    
    Note that we won't have exactly the same results as the broadening is not
    updated in MergeSlabs 
    '''
    
    from neq.spec import load_spec
    import matplotlib.pyplot as plt
    from neq.test.utils import getTestFile
    
    b = True
    
    for optically_thin in [True, False]:
    
        s1 = load_spec(getTestFile('CO_Tgas1500K_mole_fraction0.01.spec'))
        s2 = load_spec(getTestFile('CO_Tgas1500K_mole_fraction0.5.spec'))
        s1.update(optically_thin=optically_thin)
        s2.update(optically_thin=optically_thin)
    #    s1 = sf.non_eq_spectrum(Tgas, Tgas, mole_fraction=0.01)
    #    s2 = sf.eq_spectrum(Tgas, mole_fraction=N*0.01)
    #    slist = [s1.copy()]*N
        slist = [s1]*50
        s1N = MergeSlabs(*slist, optically_thin=optically_thin, debug=debug)
        
    #    print(s2.compare_with(s1N, only_spectra=True))
    #    print(np.allclose(s2.get('radiance_noslit'), s1N.get('radiance_noslit')))
    #    print('Max error: {0:.2f}%'.format(np.abs(s2.get('radiance_noslit')[1]/
    #          s1N.get('radiance_noslit')[1]-1).max()*100))
        
        b1 = np.isclose(s2.get_power(), s1N.get_power(), 1.5e-2)
        if verbose:
            print('Compare 50x[CO=0.01] vs 1x[CON=0.5] (optically thin: {1}): {0}'.format(
                    b1, optically_thin))
            print('... Difference: {0:.2f}%'.format(abs(s1N.get_power()/s2.get_power()-1)*100))
        b *= b1
    
        if plot:
            for k in ['radiance_noslit']: #, 'transmittance_noslit']:
                s2.plot(k, lw=3,label='1x[CO=0.5]')
                s1N.plot(k, nfig='same', label='50x[CO=0.01]')
                plt.legend()
                plt.title('Optically thin: {0}'.format(optically_thin))
                plt.tight_layout()
            
    return bool(b)
#        
        
def _test(verbose=True, plot=True, debug=False, warnings=True, *args, **kwargs):
    
    b = True
    
    b *= _test_merge_slabs(verbose=verbose, plot=plot, debug=debug, warnings=warnings,
                           *args, **kwargs)
    
    # Todo: add a test with _serial_slabs ... For instance the Sebastien Depraz reference case
        
    return bool(b)


if __name__ == '__main__':
    print('Testing merge slabs: ', _test(verbose=True))



# -*- coding: utf-8 -*-
"""
Created on Wed Oct 18 15:38:14 2017

@author: erwan
"""

from __future__ import print_function
from __future__ import absolute_import
import sys
from time import time
from six.moves import range

class ProgressBar():
    ''' A console progress-bar
    
    Use
    ---
    
    >>> pb = ProgressBar(N)
    >>> for i, ... in enumerate(...):
    >>>     (...)
    >>>     pb.update(i)
    >>> pb.done()
    
    See test in progress_bar.py
    
    Todo
    ----
    
    One day extend for multiprocss with several progress values? 
    https://stackoverflow.com/questions/7392779/is-it-possible-to-print-a-string-at-a-certain-screen-position-inside-idle
    
    '''
    
    def __init__(self, N, active=True):
        ''' 
        write to progress bar completion status i/N. 
        
        
        Parameters    
        ----------
        
        N: int
            total number of iterations
        
        '''
        self.t0 = time()
        self.N = N
        self.active = active
        
    def set_active(self, active=True):
        ''' Option to activate/deactivate the ProgressBar. Used not to make it 
        appear on small processes (based on a condition) without changing most
        of the code'''

        self.active = active

    def update(self, i):
        ''' 
        write to progress bar completion status i/N. If t0 is not None, also 
        write the time spent
        
        
        Parameters    
        ----------
        
        i: int
            current iteration
        '''
        
        if not self.active: return
        
        N = self.N
        t0 = self.t0
        
        if t0 is None:
            msg = '{0:.1f}%'.format(i/N*100)
        else:
            msg = '({0:.0f}s)\t{1:.1f}%'.format(time()-t0, i/N*100)
            
        if sys.stdout is not None:
            sys.stdout.write('\r'+msg)
            sys.stdout.flush()
            
    def done(self):
        
        if not self.active: return
        
        self.update(self.N)
        # make new line
        if sys.stdout is not None:
            sys.stdout.write('\n')
            sys.stdout.flush()
            
        
# %% Test     
def test_progress_bar(*args, **kwargs):
    ''' Minimal example of a progress bar '''
    
    from radis.misc.progress_bar import ProgressBar
    from time import sleep
    from numpy.random import rand
    
    print('Testing progress bar')
    
    a = 0
    r = list(range(1000))
    N = len(r)
    pb = ProgressBar(N, t0=time())
    for i in r:
        if i % 10: 
            pb.update(i)
        a += i
        sleep(rand()*3e-3)
    pb.done()
    
    return True  # nothing implemented

if __name__ == '__main__':
    test_progress_bar()


# -*- coding: utf-8 -*-
"""
Created on Sat Nov 14 19:40:53 2015

@author: Erwan

Performance test with pint
- 2 times slower than numpy on np.std(np.exp())

Import neq.phys.units to get access to:

Use
---

from radis.phys.units import Q_
- a = Q_(np.array([5,4,2]),'cm')
- a.to_base_units()

"""

from __future__ import absolute_import
from __future__ import print_function
import numpy as np
import os
from pint import UnitRegistry, DimensionalityError

ureg = UnitRegistry()
ureg.load_definitions(os.path.join(os.path.dirname(__file__), 'units.txt'))
Q_ = ureg.Quantity

# %% Pint aware arrays


class uarray(np.ndarray):
    ''' Unit-aware array based on Pint
    
    Example
    -------
    
    >>> a = uarray(np.linspace(10, 100, 10), 'Td')
    
    '''

    def __new__(cls, input_array, unit=None):
        # Input array is an already formed ndarray instance
        # We first cast to be our class type

        obj = np.asarray(input_array).view(cls)
        # add the new attribute to the created instance

        if unit is not None:
            obj = Q_(obj, unit)

        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):
        # see InfoArray.__array_finalize__ for comments
        if obj is None:
            return
#        self.info = getattr(obj, 'info', None)

def conv2(quantity, fromunit, tounit):
    ''' Converts `quantity` from unit `fromunit` to unit `tounit`
    
    
    Parameters    
    ----------
    
    quantity: array
        quantity to convert
        
    fromunit: str
        input unit
    
    tounit: str
        output unit
        
        
    Note
    ----
    
    Note that the output is still non dimensional. We don't transform `quantity` 
    into a pint array (or neq.phys.uarray) because this may create a performance
    drop in computationaly-expensive task. Instead, we assume we know for 
    sure the units in which some of our quantities will be created, and just
    want to let the users choose another output unit 
    
    '''

    try:
        a = Q_(quantity,fromunit)
        a = a.to(tounit)
    except TypeError:
        if 'cm-1' in fromunit or 'cm-1' in tounit:
            raise TypeError('Use cm_1 instead of cm-1 else it triggers errors in '+\
                            'pint (symbolic unit converter)')
        else:
            raise
            
    return a.magnitude


def is_homogeneous(unit1, unit2):
    ''' Tells if unit1 and unit2 are homogeneous, using the Pint library
    
    
    Parameters    
    ----------
    
    unit1, unit2: str
        units 

    '''
    
    try:
        Q_(unit1) + Q_(unit2)
        return True
    except DimensionalityError:
        return False
    
# %% Emission density conversion functions (with changes from ~1/nm in ~1/cm-1)

def convert_emi2cm(j_nm, wavenum, Iunit0, Iunit):
    '''
    Convert spectral emission density in wavelength base (typically ~mW/cm3/sr/nm) 
    to spectral emission density in wavenumber base (~mW/cm3/sr/cm_1)
    
    
    Parameters    
    ----------
    
    j_nm: array
        spectral emission density in Iunit0 unit (~ per wavelength) )
        
    wavenum: array (cm-1)
        wavenumber
        
    Iunit0: str
        unit (~ per wavelength) to convert from 
        
    Iunit: str
        unit (~ per wavenumber) to convert to


    Notes
    -----
    
    Implementation:

    We use the variable substitution:
            dλ/dν = - 1/ν**2        [in SI]
            dλ/dν = - 10^7 / ν**2   [nm -> cm-1]

    We want   Lν.dν = Lλ.dλ
    so        Lν = - Lλ * 1e7 / ν^2


    Example
    -------

    Validation:
    
        >>> w_nm, j_nm = s.get('emisscoeff', 'nm', 'mW/cm2/sr/nm')
        >>> w_cm, j_cm = s.get('emisscoeff', 'cm', 'mW/cm2/sr/cm_1')
        >>> print(trapz(y_nm, x_nm))
        >>> print(trapz(y_cm, x_cm))

    Both integrals are to be the same

    '''

    if Q_(Iunit0) != Q_('mW/cm3/sr/nm'):  # Q_ makes sure 'mW/sr/cm3/nm' == 'mW/cm3/sr/nm'
        j_nm = conv2(j_nm, Iunit0, 'mW/cm3/sr/nm')

    # Convert ''mW/cm3/sr/nm' to 'mW/cm3/sr/cm_1'
    j_cm = j_nm * 1e7 / wavenum**2
    # note that we discard the - sign.
    # This means that one of the integrals `trapz(j_nm, wavelen)` or
    # `trapz(j_cm, wavenum)` will be < 0 as they are sorted in opposite order.
    # But both plots look alright

    # Convert to whatever was wanted
    j_cm = conv2(j_cm, 'mW/cm3/sr/cm_1', Iunit)

    return j_cm


def convert_emi2nm(j_cm, wavenum, Iunit0, Iunit):
    '''
    Convert spectral emission density in wavenumber base (typically ~mW/cm3/sr/cm_1) to 
    spectral radiance in wavelength base (~mW/cm3/sr/nm)

    
    Parameters    
    ----------
    
    j_cm: array
        spectral emission density in Iunit0 unit (~ per wavenumber)
        
    wavenum: array (cm-1)
        wavenumber
        
    Iunit0: str
        unit (~ per wavenumber) to convert from 
        
    Iunit: str
        unit (~ per wavelength) to convert to
        
        
    Notes
    -----
    
    Implementation:

    We use the variable substitution:
            dλ/dν = - 1/ν**2        [in SI]
            dλ/dν = - 10^7 / ν**2   [nm -> cm-1]

    We want   Lν.dν = Lλ.dλ
    so        Lλ  = - Lν * 1e-7 * ν^2

    '''

    if Q_(Iunit0) != Q_('mW/cm3/sr/cm_1'):  
        j_cm = conv2(j_cm, Iunit0, 'mW/cm3/sr/cm_1')

    # Convert 'mW/cm3/sr/cm_1' to 'mW/cm3/sr/nm'
    j_nm = j_cm * 1e-7 * wavenum**2
    # note that we discard the - sign.
    # This means that one of the integrals `trapz(L_nm, wavelen)` or
    # `trapz(L_cm, wavenum)` will be < 0 as they are sorted in opposite order.
    # But both plots look alright

    # Convert to whatever was wanted
    j_nm = conv2(j_nm, 'mW/cm3/sr/nm', Iunit)

    return j_nm


# %% Radiance conversion functions (with changes from ~1/nm in ~1/cm-1)
    
def convert_rad2cm(l_nm, wavenum, Iunit0, Iunit):
    '''
    Convert spectral radiance in wavelength base (~1/nm) to spectral radiance in
    wavenumber base (~1/cm_1)
    
    
    Parameters    
    ----------
    
    l_nm: array
        spectral radiance in yunit0 unit (~ per wavelength)
        
    wavenum: array (cm-1)
        wavenumber
        
    Iunit0: str
        unit (~ per wavelength) to convert from 
        
    Iunit: str
        unit (~ per wavenumber) to convert to


    Notes
    -----
    
    Implementation:

    We use the variable substitution:
            dλ/dν = - 1/ν**2        [in SI]
            dλ/dν = - 10^7 / ν**2   [nm -> cm-1]

    We want   Lν.dν = Lλ.dλ
    so        Lν = - Lλ * 1e7 / ν^2


    Example
    -------
    
    Validation:
    
        >>> x_nm, y_nm = s.get('radiance_noslit', 'nm', 'mW/cm2/sr/nm')
        >>> x_cm, y_cm = s.get('radiance_noslit', 'cm', 'mW/cm2/sr/cm_1')
        >>> print(trapz(y_nm, x_nm))
        >>> print(trapz(y_cm, x_cm))

    Both integrals are to be the same

    '''

    if Q_(Iunit0) != Q_('mW/cm2/sr/nm'):  # Q_ makes sure 'mW/sr/cm2/nm' == 'mW/cm2/sr/nm'
        l_nm = conv2(l_nm, Iunit0, 'mW/cm2/sr/nm')

    # Convert ''mW/cm2/sr/nm' to 'mW/cm2/sr/cm_1'
    l_cm = l_nm * 1e7 / wavenum**2
    # note that we discard the - sign.
    # This means that one of the integrals `trapz(L_nm, wavelen)` or
    # `trapz(L_cm, wavenum)` will be < 0 as they are sorted in opposite order.
    # But both plots look alright

    # Convert to whatever was wanted
    l_cm = conv2(l_cm, 'mW/cm2/sr/cm_1', Iunit)

    return l_cm


def convert_rad2nm(l_cm, wavenum, Iunit0, Iunit):
    '''
    Convert spectral radiance in wavenumber base (~1/cm_1) to spectral radiance in
    wavelength base (~1/nm)

    
    Parameters    
    ----------
    
    l_cm: array
        spectral radiance in yunit0 unit (~ per wavenumber)
        
    wavenum: array (cm-1)
        wavenumber
        
    Iunit0: str
        unit (~ per wavenumber) to convert from 
        
    Iunit: str
        unit (~ per wavelength) to convert to
        
    Notes
    -----
    
    Implementation:

    We use a variable substitution:
            dλ/dν = - 1/ν**2        [in SI]
            dλ/dν = - 10^7 / ν**2   [nm -> cm-1]

    We want   Lν.dν = Lλ.dλ
    so        Lλ  = - Lν * 1e-7 * ν^2

    '''

    if Q_(Iunit0) != Q_('mW/cm2/sr/cm_1'):  # Q_ makes sure 'mW/sr/cm2/cm_1' == 'mW/cm2/sr/cm_1'
        l_cm = conv2(l_cm, Iunit0, 'mW/cm2/sr/cm_1')

    # Convert 'mW/cm2/sr/cm_1' to 'mW/cm2/sr/nm'
    l_nm = l_cm * 1e-7 * wavenum**2
    # note that we discard the - sign.
    # This means that one of the integrals `trapz(L_nm, wavelen)` or
    # `trapz(L_cm, wavenum)` will be < 0 as they are sorted in opposite order.
    # But both plots look alright

    # Convert to whatever was wanted
    l_nm = conv2(l_nm, 'mW/cm2/sr/nm', Iunit)

    return l_nm


def convert_universal(I, from_unit, to_unit, w_cm,
                     per_nm_is_like='mW/sr/cm2/nm', per_cm_is_like='mW/sr/cm2/cm_1'):
    ''' Return variable var in whatever unit, and converts to to_unit
    Also deal with cases where var is in ~1/nm (per_nm_is_like) or ~1/cm-1
    (per_cm_is_like)

    
    Parameters    
    ----------

    var: str
        variable to get. Usually 'radiance' or 'radiance_noslit'

    to_unit: str
        unit to convert variable to

    '''
    Iunit0 = from_unit
    Iunit = to_unit
    try:
        if is_homogeneous(Iunit0, per_nm_is_like) and is_homogeneous(
                Iunit, per_cm_is_like):
            I = convert_rad2cm(I, w_cm, Iunit0, Iunit)
            # note that there may still be a new DimensionalityError
            # raise here if the input was non-sense.
        elif is_homogeneous(Iunit0, per_cm_is_like) and is_homogeneous(
                Iunit, per_nm_is_like):
            I = convert_rad2nm(I, w_cm, Iunit0, Iunit)
            # note that there may still be a new DimensionalityError
            # raise here if the input was non-sense.
        else:  # general case: just convert
            I = conv2(I, Iunit0, Iunit)
    except DimensionalityError:
        raise DimensionalityError(Iunit0, Iunit)

    return I


# %% Test

def _test(verbose=True, *args, **kwargs):

    b = True

    # Test unit-ware arrays
    a = uarray(np.linspace(10, 100, 10), 'Td')          # NeQ pint-aware array
    res = Q_(np.linspace(1e-16, 1e-15, 10), 'V * cm^2')  # pint definition

    b *= (np.round(np.array(a.to('V * cm^2')) - np.array(res), 5)
         == np.zeros_like(res)).all()
#    b = (print(a.to('V * cm^2'))==print(res))


    # Test conversion 
    convtable = [
            (500, 'nm', 0.5, 'µm'),
            (1, 'erg/s', 1e-7, 'W'),
            (1, 'm2',10000,'cm2')
            ]
    for a, f, r, t in convtable:
        cr = conv2(a, f, t)
        print(('{0} {1} = {2} {3}'.format(a, f, cr, t)))
        b *= np.isclose(cr,r)

    return bool(b)

if __name__ == '__main__':

    print((_test()))



# -*- coding: utf-8 -*-
"""
Created on Thu Feb 16 14:58:40 2017

@author: erwan

"""

# TODO: remove wangl2  after loading database (wangl is enough, with 1=e and 2=f)

from __future__ import print_function, absolute_import, division, unicode_literals

import pandas as pd
import numpy as np
from collections import OrderedDict
import os
from os.path import exists, splitext
from radis.io.hitran import _cast_to_dtype, _format_dtype
import sys
from six.moves import zip

columns_hitemp = OrderedDict([ (
               # name    # format # type  # description                                 # unit 
               'id',     ('a2',   int,   'Molecular number'                               ,''                      )),(
               'iso',    ('a1',   int,   'isotope number'                                 ,''                      )),(
               'wav',    ('a12',  float, 'vacuum wavenumber'                              ,'cm-1'                  )),(
               'int',    ('a10',  float, 'intensity at 296K'                              ,'cm-1/(molecule/cm-2)', )),(
               'A',      ('a10',  float, 'Einstein A coefficient'                         ,'s-1'                   )),(
               'airbrd', ('a5',   float, 'air-broadened half-width at 296K'               ,'cm-1.atm-1'            )),(
               'selbrd', ('a5',   float, 'self-broadened half-width at 296K'              ,'cm-1.atm-1'            )),(
               'El',     ('a10',  float, 'lower-state energy'                             ,'cm-1'                  )),(
               'Tdpair', ('a4',   float, 'temperature-dependance exponent for Gamma air'  ,''                      )),(
               'Pshft',  ('a8',   float,  'air pressure-induced line shift at 296K'       ,'cm-1.atm-1'            )),(
                       # skip 1 columns   . WARNING. They didn't say that in the doc. But it doesn't make sense if i don't
               'Tdpsel', ('a5',   float, 'temperature dependance exponent for gamma self' ,''                      )),(
               'v1u',    ('a3',   int,   'upper state vibrational number v1'              ,''                      )),(
               'v2u',    ('a2',   int,   'upper state vibrational number v2'              ,''                      )),(
               'l2u',    ('a2',   int,   'upper state vibrational number l2'              ,''                      )),(
               'v3u',    ('a2',   int,   'upper state vibrational number v3'              ,''                      )),(
               'ru',     ('a1',   int,   'upper state vibrational number r'               ,''                      )),(
                       # skip 5 columns (v1l format becomes 3+5)
               'v1l',    ('a8',   int, 'lower state vibrational number v1'                ,''                      )),(
               'v2l',    ('a2',   int,   'lower state vibrational number v2'              ,''                      )),(
               'l2l',    ('a2',   int,   'lower state vibrational number l2'              ,''                      )),(
               'v3l',    ('a2',   int,   'lower state vibrational number v3'              ,''                      )),(
               'rl',     ('a1',   int,   'lower state vibrational number r'               ,''                      )),(
               'polyu',  ('a3',   int,   'upper state polyad'                             ,''                      )),(
               'wangu',  ('a2',   int,   'upper state Wang symmetry'                      ,''                      )),(
               'ranku',  ('a4',   int,   'upper state ranking number'                     ,''                      )),(
               'polyl',  ('a3',   int,   'lower state polyad'                             ,''                      )),(
               'wangl',  ('a2',   int,   'lower state Wang symmetry'                      ,''                      )),(
               'rankl',  ('a4',   int,   'lower state ranking number'                     ,''                      )),(
                       # skip 2 columns (PQR format becomes 1+2)
               'branch',    ('a3',   str,   'O, P, Q, R, S branch symbol'                                        ,''                      )),(
               'jl',     ('a3',   int,   'lower state rotational quantum number'          ,''                      )),(
               'wangl2', ('a1',   str,   'lower state Wang symmetry'                      ,'(e or f)'              )),(
               'lsrc',   ('a5',   int,   'line source'                                    ,''                      ))
                ])

columns_4000 = OrderedDict([ (
               # name    # format # type  # description                                 # unit 
               'id',     ('a2',   int,   'Molecular number'                               ,''                      )),(
               'iso',    ('a1',   int,   'isotope number'                                 ,''                      )),(
               'wav',    ('a12',  float, 'vacuum wavenumber'                              ,'cm-1'                  )),(
               'int',    ('a10',  float, 'intensity at 296K'                              ,'cm-1/(molecule/cm-2)', )),(
               'A',      ('a10',  float, 'Einstein A coefficient'                         ,'s-1'                   )),(
               'airbrd', ('a5',   float, 'air-broadened half-width at 296K'               ,'cm-1.atm-1'            )),(
               'selbrd', ('a5',   float, 'self-broadened half-width at 296K'              ,'cm-1.atm-1'            )),(
               'El',     ('a10',  float, 'lower-state energy'                             ,'cm-1'                  )),(
               'Tdpair', ('a4',   float, 'temperature-dependance exponent for Gamma air'  ,''                      )),(
               'Pshft',  ('a8',   float,  'air pressure-induced line shift at 296K'       ,'cm-1.atm-1'            )),(
                       # skip 1 columns   . WARNING. They didn't say that in the doc. But it doesn't make sense if i don't
               'Tdpsel', ('a5',   float, 'temperature dependance exponent for gamma self' ,''                      )),(
               'v1u',    ('a3',   int,   'upper state vibrational number v1'              ,''                      )),(
               'v2u',    ('a2',   int,   'upper state vibrational number v2'              ,''                      )),(
               'l2u',    ('a2',   int,   'upper state vibrational number l2'              ,''                      )),(
               'v3u',    ('a2',   int,   'upper state vibrational number v3'              ,''                      )),(
               'ru',     ('a2',   int,   'upper state vibrational number r'               ,''                      )),(
                       # skip 5 columns (v1l format becomes 3+5)
               'v1l',    ('a6',   int, 'lower state vibrational number v1'                ,''                      )),(
               'v2l',    ('a2',   int,   'lower state vibrational number v2'              ,''                      )),(
               'l2l',    ('a2',   int,   'lower state vibrational number l2'              ,''                      )),(
               'v3l',    ('a2',   int,   'lower state vibrational number v3'              ,''                      )),(
               'rl',     ('a2',   int,   'lower state vibrational number r'               ,''                      )),(
               'polyu',  ('a3',   int,   'upper state polyad'                             ,''                      )),(
               'wangu',  ('a2',   int,   'upper state Wang symmetry'                      ,''                      )),(
               'ranku',  ('a4',   int,   'upper state ranking number'                     ,''                      )),(
               'polyl',  ('a3',   int,   'lower state polyad'                             ,''                      )),(
               'wangl',  ('a2',   int,   'lower state Wang symmetry'                      ,''                      )),(
               'rankl',  ('a4',   int,   'lower state ranking number'                     ,''                      )),(
                       # skip 2 columns (PQR format becomes 1+2)
               'branch', ('a3',   str,   'O, P, Q, R, S branch symbol'                    ,''                      )),(
               'jl',     ('a3',   int,   'lower state rotational quantum number'          ,''                      )),(
               'wangl2', ('a1',   str,   'lower state Wang symmetry'                      ,'(e or f)'              ))
                ])

def cdsd2df(fname, version='hitemp', count=-1, cache=False, verbose=True):
    ''' Convert a CDSD-HITEMP file to a Pandas dataframe 
    
    
    Parameters    
    ----------
    
    fname: str
        CDSD file name 
        
    version: str ('4000', 'hitemp')
        CDSD version
        
    count: int
        number of items to read (-1 means all file)
        
    cache: boolean, or 'regen'
        if True, a pandas-readable HDF5 file is generated on first access, 
        and later used. This saves on the datatype cast and conversion and
        improves performances a lot (but changes in the database are not 
        taken into account). If False, no database is used. If 'regen', temp
        file are reconstructed. Default True. 
    
    
    Returns
    -------
    
    df: pandas Dataframe
        dataframe containing all lines and parameters
        
    
    References
    ----------
    
    CDSD-HITEMP / CDSD-4000 doc 
    
    
    Notes
    -----
    
    Performances:
    
    I had huge performance trouble with this function, because the files are 
    huge (500k lines) and the format is to special (no space between numbers...)
    to apply optimized methods such as pandas's. A line by line reading isn't
    so bad, using struct to parse each line. However, we waste typing determining
    what every line is. I ended up using the fromfiles functions from numpy,
    not considering '\n' as a special character anymore, and a second call
    to numpy to cast the correct format. That ended up being twice as fast. 
    
    - initial:                      20s / loop
    - with mmap:                    worse 
    - w/o readline().rstrip('\n'):  still 20s
    - numpy fromfiles:              17s
    - no more readline, 2x fromfile 9s
    
    Think about using cache mode too:
    - no cache mode                 9s
    - cache mode, first time        22s
    - cache mode, then              2s
    
    Moving to HDF5:
    
    On cdsd_02069_02070 (56 Mb)
    
    Reading:
        
    >>> cdsd2df(): 9.29 s
    >>> cdsd2df(cache=True [old .txt version]): 2.3s 
    >>> cdsd2df(cache=True [new h5 version, table]): 910ms
    >>> cdsd2df(cache=True [new h5 version, fixed]): 125ms
    
    Storage:
        
    >>> %timeit df.to_hdf("cdsd_02069_02070.h5", "df", format="fixed")  337ms
    >>> %timeit df.to_hdf("cdsd_02069_02070.h5", "df", format="table")  1.03s

    '''
    
    if version == 'hitemp':
        columns = columns_hitemp
    elif version == '4000':
        columns = columns_4000
    else:
        raise ValueError('Unknown CDSD version: {0}'.format(version))
    
    if cache: # lookup if cached file exist. 
#        fcache = fname+'.cached'
        fcache = splitext(fname)[0]+'.h5'
        if exists(fcache):
            if cache == 'regen':
                os.remove(fcache)
                if verbose: print('Deleted h5 cache file : {0}'.format(fcache))
            else:
                if verbose: print('Using h5 file: {0}'.format(fcache))
    #            return pd.read_csv(fcache)
                return pd.read_hdf(fcache, 'df')

    # Create a dtype with the binary data format and the desired column names
    dtype = [(k, c[0]) for (k, c) in columns.items()]+[('_linereturn','a2')]
    try:
        dt = np.dtype([(str(k), c) for k, c in dtype])
        # Note: dtype names cannot be `unicode` in Python2. Hence the str()
    except TypeError:
        # Cant read database. Try to be more explicit for user
        print('Data type')
        print('-'*30)
        for (k, c) in dtype:
            print(str(k), '\t', c)
        print('-'*30)
        raise
        
    data = np.fromfile(fname, dtype=dt, count=count)
    
    # %% Cast to new type
    
    # get format of line return
    if sys.platform in ['linux', 'darwin']:
        linereturnformat = 'a1'
    elif sys.platform in ['win32']:
        linereturnformat = 'a2'
    else:
        raise ValueError('Line return format not defined for this OS: please update neq')  
    
    newtype = [c[0] if (c[1]==str) else c[1] for c in columns.values()]
    dtype = list(zip(list(columns.keys()), newtype))+[('_linereturn',linereturnformat)]
    data2 = _cast_to_dtype(data, dtype)

    # %% Create dataframe    
    df = pd.DataFrame(data2.tolist(), columns=list(columns.keys())+['_linereturn'])
    
    for k, c in columns.items():
        if c[1] == str:
            df[k] = df[k].str.decode("utf-8")
    
    # Strip whitespaces around PQR columns (due to 2 columns jumped)
    df.branch = df.branch.str.strip()
    
    # Delete dummy column than handled the line return character
    del df['_linereturn']
    
    if cache: # cached file mode but cached file doesn't exist yet (else we had returned)
        if verbose: print('Generating cached file: {0}'.format(fcache))
        try:
#            df.to_csv(fcache)
            df.to_hdf(fcache, 'df', format='fixed')
        except:
            if verbose: print('An error occured in cache file generation. Lookup access rights')
            pass
        
    return df 


def _test(verbose=True, warnings=True, **kwargs):
    ''' Analyse some default files to make sure everything still works'''
    from neq.test.utils import getTestFile
    from time import time
    
    t0 = time()
    df = cdsd2df(getTestFile('cdsd_hitemp_09.txt'))
    print('File loaded in {0:.1f}s'.format(time()-t0))
    if verbose: print(df.head())
    return df.wav[3] == 2250.00096


if __name__ == '__main__':
    print('Testing cdsd: ', _test())



# -*- coding: utf-8 -*-
"""
Created on Thu Aug  3 15:27:36 2017

Planck functions:
- planck: planck radiation with wavelength input
- planck_wn: planck radiation with wavenumber input
- see also neq.spec.sPlanck: a NeQ Spectrum blackbody object

"""

from __future__ import absolute_import
from numpy import exp
from radis.phys.constants import k_b, c, h
from radis.phys.constants import k_b_CGS, c_CGS, h_CGS
from radis.phys.units import conv2, Q_

def planck(lmbda, T, eps=1, unit='mW/sr/cm2/nm'):
    ''' Planck function for blackbody radiation

    
    Parameters    
    ----------

    λ: np.array   (nm)
       wavelength

    T: float    (K)
        equilibrium temperature

    eps: grey-body emissivity
        default 1

    unit: output unit
        default 'mW/sr/cm2/nm'


    Returns
    -------

    planck: np.array   (mW.sr-1.cm-2/nm)
        equilibrium radiance

    '''

    k = k_b
    lbd = lmbda * 1e-9
    iplanck = eps*(2*h*c**2/lbd**5) * 1/(exp(h*c/(lbd*k*T)) - 1)   # S.I  (W.sr-1.m-3)
    iplanck *= 1e-10  # W.sr-1.m-3 >>> mW.sr-1.cm-2.nm-1

    if Q_(unit) != Q_('mW/sr/cm2/nm'):
        iplanck = conv2(iplanck, 'mW/sr/cm2/nm', unit)

    return iplanck

def planck_wn(wavenum, T, eps=1, unit='mW/sr/cm2/cm_1'):
    ''' Planck function for blackbody radiation, wavenumber version

    
    Parameters    
    ----------

    wavenum: np.array   (cm-1)
       wavenumber

    T: float    (K)
        equilibrium temperature

    eps: grey-body emissivity
        default 1

    unit: str
        output unit. Default 'mW/sr/cm2/cm_1'


    Returns
    -------

    planck: np.array   default (mW/sr/cm2/cm_1)
        equilibrium radiance

    '''

    k = k_b_CGS
    h = h_CGS
    c = c_CGS

    iplanck = eps*(2*h*c**2*wavenum**3) * 1/(exp(h*c*wavenum/(k*T)) - 1) 
    # iplanck in erg/s/sr/cm2/cm-1
    iplanck *= 1e-4     # erg/s/sr/cm2/cm-1 > mW/sr/cm^2/cm-1

    if Q_(unit) != Q_('mW/sr/cm2/cm_1'):
        iplanck = conv2(iplanck, 'mW/sr/cm2/cm_1', unit)

    return iplanck


if __name__ == '__main__':
    pass


# -*- coding: utf-8 -*-
"""
Created on Tue May 26 11:52:15 2015

Erwan Pannier
EM2C, CentraleSupélec, 2015
CNRS UPR 288

"""

from __future__ import absolute_import, division, print_function, unicode_literals

from .hitran import hit2df, get_molecule, get_molecule_identifier
from .cdsd import cdsd2df